/*
 * ultimeter.c - Ultimeter weather parser.
 * $Id: ultimeter.c,v 1.1.1.1 2003/03/25 07:14:40 augur Exp $
 *
 *   These functions were developed for APRS World.  
 *   This program serves as a simple testbed for these functions.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *  
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <string.h>
#include <stdio.h>
#include <aprsgated/metric.h>
#include <aprsgated/aprsgated.h>

int axtoi( const char *digits );
void substr (char *dest, char *src, int size) ;
/* Ultimeter data parsing functions  */
void parse_peet_packet (struct weather *wx, char *p);		/* 100,200,400,700,2000 Packet */
void parse_peet_datalog (struct weather *wx, char *p);	/* 100,200,400,700,2000 Dta Logging Mode */
void parse_peet_complete (struct weather *wx, char *p);	/* 100,200,400,700,2000 Complete Record Format */
void parse_peet_u2 (struct weather *wx, char *p);				/* U-II Packet Mode */

/* Entry Function */
int parse(struct weather *wx, char *record){
		switch( record[0] ){
			case '$':
				/* Peet Bros Packet Format */
				parse_peet_packet( wx, record );
				break;
			case '!':
				/* Peet Bros Data Logging Format */
				parse_peet_datalog( wx, record );
				break;
			case '&':
				/* Peet Bros Complete Record */
				parse_peet_complete( wx, record );
				break;
			case '*':
			case '#':
				/* Ultimeter II Record */
				parse_peet_u2( wx, record );
				break;
			default:
				/* Unknown format header. Twice in a row may mean the port is insane.  */
				return -1;
		}
		return 0;
}

void substr (char *dest, char *src, int size) {
	strncpy (dest, src, (size_t) size);
	dest[size] = '\0';
}

/* Parse Peet Bros Packet mode data.  Header is "$ULTW" */
void
parse_peet_packet (struct weather *wx, char *p) {
	char temp[5];
	
	/* Wind Peak Speed, Gust, Sustained.  Ultimeter only reports 5-min gust in Packet Mode. */
	substr(temp, p+5, 4);
	wx->wind_speed = axtoi( temp ) * 0.1;  /* Convert 1/10 KPH to KPH */

	/* Wind Peak Direction */
	substr(temp, p+9, 4);
	wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	
	/* Outdoor Temperature */
	substr(temp, p+13, 4);
	wx->temperature = F_TO_C( axtoi( temp ) * 0.1 );  /* Convert 1/10 F to C */
	
	/* Rain Total (since Ultimeter manual reset) */
	substr(temp, p+17, 4);
	wx->rain_today = IN_TO_CM( axtoi( temp ) * 0.01 );  /* Convert 1/100 IN to CM */
	
	/* Altimeter Pressure */
	substr(temp, p+21, 4);
	wx->barometer = axtoi( temp ) * 0.1;  /* Already in millibars */

	/* Percent Relative Humidity */
	substr(temp, p+37, 4);
	wx->humidity = axtoi( temp ) * 0.1;  /* Convert 1/10 rh to rh */
}

/* Parse peet bros data logging format.  Header is "!!" */
void
parse_peet_datalog (struct weather *wx, char *p) {
	char temp[5];
	
	/* Wind Peak Speed, Gust, Sustained. */
	substr(temp, p+2, 4);
	wx->wind_speed = axtoi( temp ) * 0.1;  /* Convert 1/10 KPH to KPH */

	/* Wind Peak Direction */
	substr(temp, p+6, 4);
	wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	
	/* Outdoor Temperature */
	substr(temp, p+10, 4);
	wx->temperature = F_TO_C( axtoi( temp ) * 0.1 );  /* Convert 1/10 F to C */
	
	/* Rain Total (since Ultimeter manual reset) */
	substr(temp, p+14, 4);
	wx->rain_today = IN_TO_CM( axtoi( temp ) * 0.01 );  /* Convert 1/100 IN to CM */
	
	/* Altimeter Pressure */
	substr(temp, p+18, 4);
	wx->barometer = axtoi( temp ) * 0.1;  /* Already in millibars */

	/* Percent Relative Humidity */
	substr(temp, p+26, 4);
	wx->humidity = axtoi( temp ) * 0.1;  /* Convert 1/10 rh to rh */
}

/* Parse Peet Bros complete record format.  Header is "&CR&" */
void
parse_peet_complete (struct weather *wx, char *p) {
	char temp[5];
	unsigned int tempval;
	unsigned int wind1, wind2, wind3;
	
	/* Wind Speed fields 1, 34, and 71.  Wind direction fields 2, 35, 72. */
	substr(temp, p+4, 4); wind1 = axtoi(temp);
	substr(temp, p+136, 4); wind2 = axtoi(temp);
	substr(temp, p+284, 4); wind3 = axtoi(temp);
	
	/* Select wind speed and direction based on which value is the highest. */
	/* Ugh, surely there's a way to make this pretty. */
	if( wind1 >= wind2 && wind1 >= wind3 ){
		wx->wind_speed = wind1 * 0.1;  /* Convert 1/10 KPH to KPH */
		/* Wind Direction. field 2. */
		substr(temp, p+8, 4);
		wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	}
	else if( wind2 >= wind1 && wind2 >= wind3 ){
		wx->wind_speed = wind2 * 0.1;  /* Convert 1/10 KPH to KPH */
		/* Wind Direction. field 35. */
		substr(temp, p+140, 4);
		wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	}
	else if( wind3 >= wind2 && wind3 >= wind1 ){
		wx->wind_speed = wind3 * 0.1;  /* Convert 1/10 KPH to KPH */
		/* Wind Direction. field 72. */
		substr(temp, p+288, 4);
		wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	}
	else{   /* Just default to the first value rather than burn more CPU time. */
		wx->wind_speed = wind1 * 0.1;  /* Convert 1/10 KPH to KPH */
		/* Wind Direction. field 2. */
		substr(temp, p+8, 4);
		wx->wind_direction = axtoi( temp ) * 1.41176;  /* Convert 1/255 into 1/360 */
	}

	/* Outdoor Temperature. field 6 */
	substr(temp, p+24, 4);
	wx->temperature = F_TO_C( axtoi( temp ) * 0.1 );  /* Convert 1/10 F to C */
	
	/* Rain Total since midnight. field 7. */
	substr(temp, p+28, 4);
	wx->rain_today = IN_TO_CM( axtoi( temp ) * 0.01 );  /* Convert 1/100 IN to CM */
	
	/* Altimeter Pressure. field 8 */
	substr(temp, p+32, 4);
	wx->barometer = axtoi( temp ) * 0.1;  /* Already in millibars */

	/* Indoor Temperature. field 12 */
	substr(temp, p+48, 4);
	wx->in_temperature = F_TO_C( axtoi( temp ) * 0.1 );  /* Convert 1/10 F to C */
	
	/* Outdoor Percent Relative Humidity.  field 13 */
	substr(temp, p+52, 4);
	wx->humidity = axtoi( temp ) * 0.1;  /* Convert 1/10 rh to rh */
	
	/* Indoor Percent Relative Humidity.  field 13 */
	substr(temp, p+56, 4);
	wx->in_humidity = axtoi( temp ) * 0.1;  /* Convert 1/10 rh to rh */
}

/* Parse Peet Bros Packet mode data.  Header is "$ULTW" */
void
parse_peet_u2 (struct weather *wx, char *p) {
	char temp[5];
	unsigned int metric;
	/*
		*or # START OF RECORD (1BYTE)
		CURRENT WIND DIRECTION (1 HEX DIGIT) 0 IS NORTH, 4 IS EAST, ETC.
		CURRENT WIND SPEEED (2 HEX DIGITS) IN MPH IF *: IN KMPH IF #
		CURRENT TMEPERATURE (2 HEX DIGITS) SUBTRACT 56 FOR DEGREES f
		UPPER RAIN GAUGE (4 HEX DIGITS) DIVIDE BY 10 OR 100 FOR INCHES*
		LOWER RAIN GAUGE (4 HEX DIGITS) DIVIDE BY 10 OR 100 FOR  INCHES*
	*/

	/*  Metric or US Measurements */
	substr(temp, p, 1);
	if( !strcmp( temp, "#" ) ){
		metric = 1;
	}
	else if( !strcmp( temp, "*" ) ){
		metric = 0;
	}
	
	/* Wind Peak Direction */
	substr(temp, p+1, 1);
	wx->wind_direction = axtoi( temp ) * 22.5;  /* Convert 1/16 into 1/360 */

	/* Wind Peak Speed, Gust, Sustained.  Ultimeter only reports 5-min gust in Packet Mode. */
	substr(temp, p+2, 2);
	if( metric == 1 ){
		wx->wind_speed = axtoi( temp );  /* KPH */
	}
	else {
		wx->wind_speed = MILE_TO_KM( axtoi( temp ) );  /* MPH to KPH */
	}
	
	/* Outdoor Temperature */
	substr(temp, p+4, 2);
	wx->temperature = F_TO_C( axtoi( temp ) - 56 );  /* Convert ( F + 56 ) to C */
	
	/* Rain total since midnight */
	substr(temp, p+10, 4);
	wx->rain_today = IN_TO_CM( axtoi( temp ) * 0.01 );  /* Convert 1/10 IN to CM */

}


/* ASCII Hexadecimal to integer
** Similar to atoi() but assumes Hexadecimal (base-16)
*/
int
axtoi( const char *digits ){
	unsigned int retval,i;
	i = retval = 0;
	
	/* While we have a digit 0-9 or A-F.  Meaning, we break on the first non-hexadecimal character just like atoi. */
	while( (digits[i] >= '0' && digits[i] <= '9') || (digits[i] >= 'A' && digits[i] <= 'F') ){
		/* Multiply value by 16. No effect if this is our first pass */
		retval = retval * 16;

		/* Convert ASCII into Integer by subtracting either 0x37 from A-F or 0x30 from 0-9 */
		retval += (unsigned int)( (digits[i] > '9') ? digits[i] - 0x37 : digits[i] - 0x30 );

		i++;
	}
	
	return retval;
}




