/* vim:ts=4:
APRS Parser with Database Support
Copyright (C) 2008 James Jefferson Jarvis	<kb0thn@aprsworld.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* I think this motto was in mind when APRS was designed:
 * "Never make anything simple and efficient when a way can be found to make it
 * complex and wonderful" -- Found in the fortune
 */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <time.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <signal.h>
#include <errno.h>
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_WINSOCK_H
/* In case of insanity break glass. */
#include <winsock.h>
#endif
#include <math.h>
#include <pwd.h>
#include <libpq-fe.h>
#include "db_pgsql.h"
#include "message.h"
#include "sqlinsert.h"

int log_postgres_error() {
	int foo = 1;

	openlog("APRS Parser (PostgreSQL)",LOG_PERROR,LOG_USER);
	syslog(LOG_NOTICE,"PostgreSQL error: %d, %s:\n%s\n", PQresultStatus(pg_res), PQerrorMessage(pg_conn), sql_buf);
	closelog();

	db_rollback_txn();
	db_begin_txn();
}


int db_begin_txn() {
	if (get_verbose() & VERBOSE_SQL)
		printf("SQL: begin\n");


	sprintf(sql_buf, "begin work");

	pg_res = PQexec(pg_conn, sql_buf);
				
	if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
		PQclear(pg_res);
		log_postgres_error();
		return -1;
	}

	PQclear(pg_res);
	return 0;
}



int db_commit_txn() {
	if (get_verbose() & VERBOSE_SQL)
		printf("SQL: commit\n");

	sprintf(sql_buf, "commit");

	pg_res = PQexec(pg_conn, sql_buf);
				
	if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
		PQclear(pg_res);
		log_postgres_error();
		return -1;
	}

	PQclear(pg_res);
	return 0;
}

int db_rollback_txn() {
	if (get_verbose() & VERBOSE_SQL)
		printf("SQL: rollback\n");

	sprintf(sql_buf, "rollback");

	pg_res = PQexec(pg_conn, sql_buf);
				
	if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
		PQclear(pg_res);
		log_postgres_error();
		return -1;
	}

	PQclear(pg_res);
	return 0;
}



char *postgres_escape(char *str, size_t length, size_t *newlen) {
	if (pg_escape_buf_index >= PG_ESCAPE_BUF_SIZE - 1) {
		pg_escape_buf_index = 0;
	} else {
		++pg_escape_buf_index;
	}

	*newlen = PQescapeStringConn(pg_conn, pg_escape_buf[pg_escape_buf_index], str, length, NULL);

	if (*newlen > PG_ESCAPE_BUF_STRING_SIZE) {
		fprintf(stderr, "Overflow quoting.\n");
		exit(1);
	}

	return pg_escape_buf[pg_escape_buf_index];
}


int make_connection (char *host, char *username, char *password, char *database) {
	int i;

	char conninfo[1024];

	if (strlen(password) > 0) {
		sprintf(conninfo, "dbname=%s user=%s password=%s host=%s", database, username, password, host);
	} else {
		sprintf(conninfo, "dbname=%s user=%s host=%s", database, username, host);
	}

    if( get_verbose() & VERBOSE_DEBUG ){
        printf("%s\n",conninfo);
    }
    
	pg_conn = PQconnectdb(conninfo);

	if (PQstatus(pg_conn) != CONNECTION_OK) {
		return -1;
	} else {

		for(i = 0; i < PG_ESCAPE_BUF_SIZE; ++i) {
			pg_escape_buf[i] = calloc(1, PG_ESCAPE_BUF_STRING_SIZE);
		}

		return 0;
	}
}

void close_connection(void) {
    PQfinish(pg_conn);
}

unsigned long packet_insert (struct raw_packet *raw) {
	struct sql_insert sql_query;
	unsigned long packet_id = 1;
	int newlen;

		/*
		 * Now that we've got a packet ID, insert the data.
		*/

		if ( ( get_insert_information()) &&
			 ( get_insert_mask() & INSERT_RAW ) ) {
			// Insert raw data
			snprintf(sql_buf, 4096, "\
				INSERT INTO aprs_raw (\
					source, \
					destination,\
					digipeater_0,\
					digipeater_1,\
					digipeater_2,\
					digipeater_3,\
					digipeater_4,\
					digipeater_5,\
					digipeater_6,\
					digipeater_7,\
					information,\
					hostid\
				) VALUES (\
					'%s', \
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					'%s',\
					%d\
				)", 

				postgres_escape(raw->source, strlen(raw->source), &newlen), 
				postgres_escape(raw->destination, strlen(raw->destination), &newlen), 
				postgres_escape(raw->digipeater[0], strlen(raw->digipeater[0]), &newlen), 
				postgres_escape(raw->digipeater[1], strlen(raw->digipeater[1]), &newlen), 
				postgres_escape(raw->digipeater[2], strlen(raw->digipeater[2]), &newlen), 
				postgres_escape(raw->digipeater[3], strlen(raw->digipeater[3]), &newlen), 
				postgres_escape(raw->digipeater[4], strlen(raw->digipeater[4]), &newlen), 
				postgres_escape(raw->digipeater[5], strlen(raw->digipeater[5]), &newlen), 
				postgres_escape(raw->digipeater[6], strlen(raw->digipeater[6]), &newlen), 
				postgres_escape(raw->digipeater[7], strlen(raw->digipeater[7]), &newlen), 
				postgres_escape(raw->information, strlen(raw->information), &newlen), 
				qa_hostid
			);

			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);
				
			if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
				return -1;
			}

			PQclear(pg_res);
		} else {
			// do not insert raw data
			// We must increment the aprs_raw_seq number so we can insert to other tables
			snprintf(sql_buf, 4096, "SELECT nextval('aprs_raw_seq')");

			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);
				
			if (PQresultStatus(pg_res) != PGRES_TUPLES_OK) {
				PQclear(pg_res);
				log_postgres_error();
				return -1;
			}

			PQclear(pg_res);
		}


		return packet_id;
}


void message_insert (struct message *msg, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen;
    
		if ( ! (get_insert_mask()&INSERT_MESSAGE) )
			return;

		snprintf(sql_buf, 4096, "\
			INSERT INTO aprs_messages (\
				source,\
				addressee,\
				message,\
				id\
			) VALUES (\
				'%s', \
				'%s', \
				'%s', \
				'%s'\
			)", 

			postgres_escape(msg->source, strlen(msg->source), &newlen), 
			postgres_escape(msg->addressee, strlen(msg->addressee), &newlen), 
			postgres_escape(msg->text, strlen(msg->text), &newlen), 
			postgres_escape(msg->id, strlen(msg->id), &newlen)
			);

		if (get_verbose() & VERBOSE_SQL)
			printf("SQL: %s\n", sql_buf);

		pg_res = PQexec(pg_conn, sql_buf);
				
		if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
			PQclear(pg_res);
			log_postgres_error();
		} else {
			PQclear(pg_res);
		}
}

void status_insert (struct status *stat, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen;

	if ( ! (get_insert_mask()&INSERT_STATUS) )
		return;

    snprintf(sql_buf, 4096, "\
        INSERT INTO aprs_status (\
            source,\
            power,\
            height,\
            gain,\
            directivity,\
            rate,\
            comment,\
            symbol_table,\
            symbol_code\
        ) VALUES (\
            '%s', \
            %d, \
            %d, \
            %lf, \
            %d, \
            %d, \
            '%s',\
            '%s', \
            '%s' \
        )", 

        postgres_escape(stat->source, strlen(stat->source), &newlen), 
        stat->power,
        stat->height,
        stat->gain,
        stat->directivity,
        stat->rate,
        postgres_escape(stat->comment, strlen(stat->comment), &newlen), 
        postgres_escape(&stat->symbol_table, 1, &newlen), 
        postgres_escape(&stat->symbol_code, 1, &newlen)
        );

    if (get_verbose() & VERBOSE_SQL)
        printf("SQL: %s\n", sql_buf);

    pg_res = PQexec(pg_conn, sql_buf);
            
    if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
        PQclear(pg_res);
        log_postgres_error();
    } else {
        PQclear(pg_res);
    }
}



void laststatus_insert (struct status *stat, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen;

		if ( ( get_do_last_tables() == 1 ) &&
			 ( get_insert_mask() & INSERT_LAST ) ) {

			snprintf(sql_buf, 4096, "delete from aprs_laststatus where source = '%s'", postgres_escape(stat->source, strlen(stat->source), &newlen));
			PQclear(PQexec(pg_conn, sql_buf));

			snprintf(sql_buf, 4096, "\
				INSERT INTO aprs_laststatus (\
					source,\
					power,\
					height,\
					gain,\
					directivity,\
					rate,\
					comment,\
					symbol_table,\
					symbol_code\
				) VALUES (\
					'%s', \
					%d,\
					%d, \
					%lf, \
					%d,\
					%d,\
					'%s',\
					'%s',\
					'%s'\
				)", 

				postgres_escape(stat->source, strlen(stat->source), &newlen), 
				stat->power,
				stat->height,
				stat->gain,
				stat->directivity,
				stat->rate,
				postgres_escape(stat->comment, strlen(stat->comment), &newlen), 
				postgres_escape(&stat->symbol_table, 1, &newlen), 
				postgres_escape(&stat->symbol_code, 1, &newlen)
				);

			

			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);
				
			if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
			} else {
				PQclear(pg_res);	
			}
		}
}


void position_insert (struct position *pos, unsigned long packet_id) {
	struct sql_insert sql_query;
	char *quotebuf[1024];
	size_t newlen;

	if ( ! (get_insert_mask() & INSERT_POSITION) )
		return;

	/* We don't even want to bother inserting if the position looks to be obviously bogus.
	* This gets rid of unlocked GPSs, unconfigured APRS clients, and the like.  
 	*/ 

	if ( pos->latitude != 0 && pos->longitude != 0 ) {	

			snprintf(sql_buf, 1024, "\
				INSERT INTO aprs_position (\
					source, \
					time_of_fix, \
					course, \
					speed, \
					altitude, \
					position\
				) VALUES (\
					'%s', \
					%ld,\
					%f, \
					%f, \
					%f, \
					GeometryFromText('POINT(%f %f)', -1)\
				)", 

				postgres_escape(pos->source, strlen(pos->source), &newlen), 
				pos->time_of_fix, 
				pos->course_made_good, 
				pos->speed_over_ground, 
				pos->altitude, 
				pos->longitude, 
				pos->latitude);

			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);
				
			if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
			} else {
				PQclear(pg_res);
			}
	}
}


void lastposition_insert (struct position *pos, struct status *stat, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen, count;

	if ( ! (get_insert_mask() & INSERT_LAST) )
		return;

	/* We don't even want to bother inserting if the position looks to be obviously bogus.
 	* This gets rid of unlocked GPSs, unconfigured APRS clients, and the like.  
 	*/ 

	if ( pos->latitude != 0 && pos->longitude != 0 ) {	

			if (get_do_last_tables() == 1) {

				/* Update existing lastposition row */

				snprintf(sql_buf, 4096, "\
					UPDATE aprs_lastposition  SET\
						packet_date = now(),\
						course = %f,\
						speed = %f,\
						altitude = %f,\
						status = '%s',\
						symbol_table = '%s',\
						symbol_code = '%s',\
						position = GeometryFromText('POINT(%f %f)', -1)\
					WHERE\
						source = '%s'\
					",

					pos->course_made_good, 
					pos->speed_over_ground, 
					pos->altitude, 
					postgres_escape(stat->comment, strlen(stat->comment), &newlen), 
					postgres_escape(&stat->symbol_table, 1, &newlen), 
					postgres_escape(&stat->symbol_code, 1, &newlen), 
					pos->longitude,
					pos->latitude, 
					postgres_escape(pos->source, strlen(pos->source), &newlen)
					);
	
				if (get_verbose() & VERBOSE_SQL)
					printf("SQL: %s\n", sql_buf);

				pg_res = PQexec(pg_conn, sql_buf);
					
				if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
					PQclear(pg_res);
					log_postgres_error();
				} else {
					count = atoi(PQcmdTuples(pg_res));
					PQclear(pg_res);
				}


				/*
				 * If the update affected zero rows, then there was no previous position report to update,
				 * and so we need to add the row.
				 */
				if (count == 0) {

					snprintf(sql_buf, 4096, "\
						INSERT INTO aprs_lastposition (\
							source,\
							course,\
							speed,\
							altitude,\
							status,\
							symbol_table,\
							symbol_code,\
							position\
						) VALUES (\
							'%s', \
							%f, \
							%f, \
							%f, \
							'%s',\
							'%s',\
							'%s',\
							GeometryFromText('POINT(%f %f)', -1)\
						)", 
	
						postgres_escape(pos->source, strlen(pos->source), &newlen), 
						pos->course_made_good, 
						pos->speed_over_ground, 
						pos->altitude, 
						postgres_escape(stat->comment, strlen(stat->comment), &newlen), 
						postgres_escape(&stat->symbol_table, 1, &newlen), 
						postgres_escape(&stat->symbol_code, 1, &newlen), 
						pos->longitude,
						pos->latitude);
	
					if (get_verbose() & VERBOSE_SQL)
						printf("SQL: %s\n", sql_buf);

					pg_res = PQexec(pg_conn, sql_buf);
				
					if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
						PQclear(pg_res);
						log_postgres_error();
					} else {
						PQclear(pg_res);
					}

				}
			}
        }
}

/*
 * Returns:
 *   hostid on success
 *   0 on failure
 *
 */
int set_qa_hostid( char *hostname ) {
	int retval = 0;
	int assign_new_host=0;
	int newlen;
    
		snprintf(sql_buf, 512, "select id from aprs_qa_hosts where host = '%s' LIMIT 1", postgres_escape(hostname, strlen(hostname), &newlen));

		pg_res = PQexec(pg_conn, sql_buf);

		if (PQresultStatus(pg_res) != PGRES_TUPLES_OK && PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
			PQclear(pg_res);
			log_postgres_error();
			return -1;
		}

		if (PQntuples(pg_res) > 0) {
			retval = atoi(PQgetvalue(pg_res, 0, 0));
			PQclear(pg_res);


		} else {
			PQclear(pg_res);

			/* Get the next ID */

			sprintf(sql_buf, "select nextval('aprs_qa_hosts_seq')");
			pg_res = PQexec(pg_conn, sql_buf);

			if (PQresultStatus(pg_res) != PGRES_TUPLES_OK && PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
				return -1;
			}

			retval = atoi(PQgetvalue(pg_res, 0, 0));

			PQclear(pg_res);


			/* Insert the host row */

			snprintf(sql_buf, 512, "\
				INSERT INTO aprs_qa_hosts (\
					id,\
					host\
				) VALUES (\
					%d, \
					'%s'\
				)", 

				retval, 
				postgres_escape(hostname, strlen(hostname), &newlen));


			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);
				
			if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
				return -1;
			}

			PQclear(pg_res);

		}

	return retval;
}


void watch_trigger(char *callsign, unsigned long packet_id) {
    /* Yikes. I don't even want to touch this. Not yet. -eburrows */
    return;
}


void weather_insert (struct weather *wx, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen, count;

	if ( ! (get_insert_mask() & INSERT_WEATHER) )
		return;

    snprintf(sql_buf, 4096, "\
        INSERT INTO aprs_weather (\
            source,\
            wind_direction,\
            wind_speed,\
            wind_gust,\
            wind_sustained,\
            temperature,\
            rain_hour,\
            rain_calendar_day,\
            rain_24hour_day,\
            humidity,\
            barometer,\
            luminosity\
        ) VALUES (\
            '%s',\
            %d,\
            %d,\
            %d,\
            %d,\
            %d,\
            %f,\
            %f,\
            %f,\
            %d,\
            %f,\
            %d\
        )", 

        postgres_escape(wx->source, strlen(wx->source), &newlen),
        wx->wind_direction,
        wx->wind_speed,
        wx->wind_gust,
        wx->wind_sustained,
        wx->temperature,
        wx->rain_hour,
        wx->rain_calendar_day,
        wx->rain_24hour_day,
        wx->humidity,
        wx->barometer,
        wx->luminosity
        );

    if (get_verbose() & VERBOSE_SQL)
        printf("SQL: %s\n", sql_buf);

    pg_res = PQexec(pg_conn, sql_buf);
            
    if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
        PQclear(pg_res);
        log_postgres_error();
    } else {
        PQclear(pg_res);
    }

    /* 
     * Update aprs_lastweather too 
     */

    if ( ( get_do_last_tables() == 1 ) &&
		 ( get_insert_mask() & INSERT_LAST ) ) {

        snprintf(sql_buf, 4096, "\
            UPDATE aprs_lastweather SET\
                packet_date = now(),\
                wind_direction = %d,\
                wind_speed = %d,\
                wind_gust = %d,\
                wind_sustained = %d,\
                temperature = %d,\
                rain_hour = %f,\
                rain_calendar_day = %f,\
                rain_24hour_day = %f,\
                humidity = %d,\
                barometer = %f,\
                luminosity = %d\
            WHERE\
                source = '%s'\
            ", 

            wx->wind_direction,
            wx->wind_speed,
            wx->wind_gust,
            wx->wind_sustained,
            wx->temperature,
            wx->rain_hour,
            wx->rain_calendar_day,
            wx->rain_24hour_day,
            wx->humidity,
            wx->barometer,
            wx->luminosity,
            postgres_escape(wx->source, strlen(wx->source), &newlen)
            );

        if (get_verbose() & VERBOSE_SQL)
            printf("SQL: %s\n", sql_buf);

        pg_res = PQexec(pg_conn, sql_buf);
                
        if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
            PQclear(pg_res);
            log_postgres_error();
        } else {
            count = atoi(PQcmdTuples(pg_res));
            PQclear(pg_res);
        }

        // If the update affected zero rows, insert the new record.
        if (count == 0) {
            snprintf(sql_buf, 4096, "\
                INSERT INTO aprs_lastweather (\
                    source,\
                    wind_direction,\
                    wind_speed,\
                    wind_gust,\
                    wind_sustained,\
                    temperature,\
                    rain_hour,\
                    rain_calendar_day,\
                    rain_24hour_day,\
                    humidity,\
                    barometer,\
                    luminosity\
                ) VALUES (\
                    '%s',\
                    %d,\
                    %d,\
                    %d,\
                    %d,\
                    %d,\
                    %f,\
                    %f,\
                    %f,\
                    %d,\
                    %f,\
                    %d\
                )",

                postgres_escape(wx->source, strlen(wx->source), &newlen),
                wx->wind_direction,
                wx->wind_speed,
                wx->wind_gust,
                wx->wind_sustained,
                wx->temperature,
                wx->rain_hour,
                wx->rain_calendar_day,
                wx->rain_24hour_day,
                wx->humidity,
                wx->barometer,
                wx->luminosity
                );

            if (get_verbose() & VERBOSE_SQL)
                printf("SQL: %s\n", sql_buf);

            pg_res = PQexec(pg_conn, sql_buf);
            
            if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
                PQclear(pg_res);
                log_postgres_error();
            } else {
                PQclear(pg_res);
            }
        }
    }
}

void last_user_insert (char *source, char *information, unsigned long packet_id) {
	int i;
	struct sql_insert sql_query;
	int newlen, count;

	if ( ( get_do_last_tables() == 1 ) &&
		 ( get_insert_mask() & INSERT_LAST ) ) {
		/* Update existing lastuserdefined row */
		snprintf(sql_buf,4096,"\
			UPDATE aprs_lastuserdefined SET\
				packet_date = now(), \
				user_id = '%s',\
				packet_type = '%s',\
				information = '%s'\
			WHERE \
				source = '%s' \
			",

			postgres_escape(&information[1],1,&newlen),
			postgres_escape(&information[2],1,&newlen),
			postgres_escape(information,strlen(information),&newlen),
			postgres_escape(source,strlen(source),&newlen)
		);

		if (get_verbose() & VERBOSE_SQL)
			printf("SQL: %s\n", sql_buf);

		pg_res = PQexec(pg_conn, sql_buf);
					
		if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
			PQclear(pg_res);
			log_postgres_error();
		} else {
			count = atoi(PQcmdTuples(pg_res));
			PQclear(pg_res);
		}


		/*
		 * If the update affected zero rows, then there was no previous position report to update,
		 * and so we need to add the row.
		 */
		if (count == 0) {
			snprintf(sql_buf,4096,"\
				INSERT INTO aprs_lastuserdefined (\
					source, \
					packet_date, \
					user_id, \
					packet_type, \
					information \
				) VALUES ( \
					'%s', \
					now(), \
					'%s',\
					'%s',\
					'%s'\
				)",

				postgres_escape(source,strlen(source),&newlen),
				postgres_escape(&information[1],1,&newlen),
				postgres_escape(&information[2],1,&newlen),
				postgres_escape(information,strlen(information),&newlen)
			);

			if (get_verbose() & VERBOSE_SQL)
				printf("SQL: %s\n", sql_buf);

			pg_res = PQexec(pg_conn, sql_buf);

			if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
				PQclear(pg_res);
				log_postgres_error();
			} else {
				PQclear(pg_res);
			}
		}
	}
}

void user_insert (char *source, char *information, unsigned long packet_id) {
	int i;
	struct sql_insert sql_query;
	int newlen;

	if ( ! (get_insert_mask()&INSERT_USER) )
		return;

	last_user_insert(source,information,packet_id);

	snprintf(sql_buf,4096,"\
		INSERT INTO aprs_userdefined (\
			source, \
			packet_date, \
			user_id, \
			packet_type, \
			information \
		) VALUES ( \
			'%s', \
			now(), \
			'%s',\
			'%s',\
			'%s'\
		)",

		postgres_escape(source,strlen(source),&newlen),
		postgres_escape(&information[1],1,&newlen),
		postgres_escape(&information[2],1,&newlen),
		postgres_escape(information,strlen(information),&newlen)
	);

	if (get_verbose() & VERBOSE_SQL)
		printf("SQL: %s\n", sql_buf);

	pg_res = PQexec(pg_conn, sql_buf);

	if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
		PQclear(pg_res);
		log_postgres_error();
	} else {
		PQclear(pg_res);
	}
}

void telemetry_insert (struct telemetry *telem, unsigned long packet_id) {
	int i;
	struct sql_insert sql_query;
	int newlen;

	if ( ! (get_insert_mask()&INSERT_TELEM) ) 
		return;

    snprintf(sql_buf, 4096, "\
        INSERT INTO aprs_telemetry (\
            source,\
            sequence,\
            analog_0,\
            analog_1,\
            analog_2,\
            analog_3,\
            analog_4,\
            digital\
        ) VALUES (\
            '%s',\
            %d, \
            %d, \
            %d, \
            %d, \
            %d, \
            %d, \
            %d\
        )",

        postgres_escape(telem->source, strlen(telem->source), &newlen), 
        telem->sequence,
        telem->analog[0],
        telem->analog[1],
        telem->analog[2],
        telem->analog[3],
        telem->analog[4],
        telem->digital
        );

    if (get_verbose() & VERBOSE_SQL)
        printf("SQL: %s\n", sql_buf);

    pg_res = PQexec(pg_conn, sql_buf);
            
    if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
        PQclear(pg_res);
        log_postgres_error();
    } else {
        PQclear(pg_res);
    }
}

void telemetry_labels_units(char *p, char *source, char *table) {
	/* UNIT.volt,deg.C,None,None,None,high,high,hig,hig,hig,hi,hi,hi */
	int i;
	char *val;
	struct sql_insert sql_query;
	int newlen;
	char *tmptok;

	char a_0[128];
	char a_1[128];
	char a_2[128];
	char a_3[128];
	char a_4[128];

	char d_0[128];
	char d_1[128];
	char d_2[128];
	char d_3[128];
	char d_4[128];
	char d_5[128];
	char d_6[128];
	char d_7[128];

		if ( ! (get_insert_mask()&INSERT_TELEM) ) 
			return;

		tmptok = strtok(p, ",");
		if (tmptok != NULL)
			strncpy(a_0, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(a_1, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(a_2, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(a_3, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(a_4, tmptok, 128);
	

		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_0, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_1, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_2, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_3, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_4, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_5, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_6, tmptok, 128);
	
		tmptok = strtok(NULL, ",");
		if (tmptok != NULL)
			strncpy(d_7, tmptok, 128);
	

		snprintf(sql_buf, 4096, "\
			INSERT INTO %s%s (\
				source, \
				a_0,\
				a_1,\
				a_2,\
				a_3,\
				a_4,\
				d_0,\
				d_1,\
				d_2,\
				d_3,\
				d_4,\
				d_5,\
				d_6,\
				d_7\
			) VALUES (\
				'%s', \
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s',\
				'%s'\
			)", 

			"aprs_",
			table,
			postgres_escape(source, strlen(source), &newlen), 
			postgres_escape(a_0, strlen(a_0), &newlen), 
			postgres_escape(a_1, strlen(a_1), &newlen), 
			postgres_escape(a_2, strlen(a_2), &newlen), 
			postgres_escape(a_3, strlen(a_3), &newlen), 
			postgres_escape(a_4, strlen(a_4), &newlen), 

			postgres_escape(d_0, strlen(d_0), &newlen), 
			postgres_escape(d_1, strlen(d_1), &newlen), 
			postgres_escape(d_2, strlen(d_2), &newlen), 
			postgres_escape(d_3, strlen(d_3), &newlen), 
			postgres_escape(d_4, strlen(d_4), &newlen), 
			postgres_escape(d_5, strlen(d_5), &newlen), 
			postgres_escape(d_6, strlen(d_6), &newlen), 
			postgres_escape(d_7, strlen(d_7), &newlen)
			);

		if (get_verbose() & VERBOSE_SQL)
			printf("SQL: %s\n", sql_buf);

		pg_res = PQexec(pg_conn, sql_buf);
				
		if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
			PQclear(pg_res);
			log_postgres_error();
		} else {
			PQclear(pg_res);
		}
}


/* should handle this INSERT in a way that doesn't duplicate coefficients calc */
void telemetry_coefficients(char *p, char *source) {
	int i;
	char *val;
	int newlen;

	double a[5];
	double b[5];
	double c[5];

	for ( i=0 ; i<5 ; i++ ) {
		a[i]=0;
		b[i]=1;
		c[i]=0;
	}

	if ( strlen(p) < 5 )
		return;

	p += 5;	// swallow the EQNS.

	i=0;
	val = strtok(p,",");
	do {
		if ( 0 == val )
			break;
		a[i] = atof(val);
		val = strtok(NULL,",");
		if ( 0 == val )
			break;
		b[i] = atof(val);
		val = strtok(NULL,",");
		if ( 0 == val )
			break;
		c[i] = atof(val);

		i++;
		val = strtok(NULL,",");
	} while ( i<5 );

	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("Telemetry coefficients:\n");
		for ( i=0 ; i<5 ; i++ ) {
			printf("\t[%d] a=%lf b=%lf c=%lf\n",i,a[i],b[i],c[i]);
		}
	}

		snprintf(sql_buf, 4096, "\
			INSERT INTO aprs_telemetry_coefficients (\
				source, \
				a_0,\
				b_0,\
				c_0,\
				a_1,\
				b_1,\
				c_1,\
				a_2,\
				b_2,\
				c_2,\
				a_3,\
				b_3,\
				c_3,\
				a_4,\
				b_4,\
				c_4\
			) VALUES (\
				'%s', \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf, \
				%lf\
			)",

			postgres_escape(source, strlen(source), &newlen), 
			a[0],b[0],c[0],
			a[1],b[1],c[1],
			a[2],b[2],c[2],
			a[3],b[3],c[3],
			a[4],b[4],c[4]);

		if (get_verbose() & VERBOSE_SQL)
			printf("SQL: %sql_buf\n", sql_buf);

		pg_res = PQexec(pg_conn, sql_buf);
				
		if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
			PQclear(pg_res);
			log_postgres_error();
		} else {
			PQclear(pg_res);
		}
}

void object_insert (struct object *obj, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen, count;

    snprintf(sql_buf, 4096, "\
        INSERT INTO aprs_object (\
            source,\
            name,\
            wind_direction,\
            wind_speed,\
            wind_gust,\
            wind_sustained,\
            temperature,\
            rain_hour,\
            rain_calendar_day,\
            rain_24hour_day,\
            humidity,\
            barometer,\
            luminosity,\
            symbol_table,\
            symbol_code,\
            comment,\
            course,\
            speed,\
            altitude,\
            position\
        ) VALUES (\
            '%s',\
            '%s',\
            %d,\
            %d,\
            %d,\
            %d,\
            %d,\
            %f,\
            %f,\
            %f,\
            %d,\
            %f,\
            %d,\
            '%s',\
            '%s',\
            '%s',\
            %f,\
            %f,\
            %f,\
            GeometryFromText('POINT(%f %f)', -1)\
        )",

        postgres_escape(obj->source, strlen(obj->source), &newlen),
        postgres_escape(obj->name, strlen(obj->name), &newlen),
        obj->wx.wind_direction,
        obj->wx.wind_speed,
        obj->wx.wind_gust,
        obj->wx.wind_sustained,
        obj->wx.temperature,
        obj->wx.rain_hour,
        obj->wx.rain_calendar_day,
        obj->wx.rain_24hour_day,
        obj->wx.humidity,
        obj->wx.barometer,
        obj->wx.luminosity,
        postgres_escape(&obj->stat.symbol_table, 1, &newlen), 
        postgres_escape(&obj->stat.symbol_code, 1, &newlen),
        postgres_escape(obj->stat.comment, strlen(obj->stat.comment), &newlen), 
        obj->pos.course_made_good, 
        obj->pos.speed_over_ground, 
        obj->pos.altitude, 
        obj->pos.longitude,
        obj->pos.latitude
        );

    if (get_verbose() & VERBOSE_SQL)
        printf("SQL: %s\n", sql_buf);

    pg_res = PQexec(pg_conn, sql_buf);
            
    if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
        PQclear(pg_res);
        log_postgres_error();
    } else {
        PQclear(pg_res);
    }


    if (obj->kill == 1) {

        snprintf(sql_buf, 4096, "delete from aprs_lastobject where source = '%s' and name = '%s'", postgres_escape(obj->source, strlen(obj->source), &newlen), postgres_escape(obj->name, strlen(obj->name), &newlen));

        if (get_verbose() & VERBOSE_SQL) {
            printf("---> Killing object %s\n", obj->name);
            printf("SQL: %s\n", sql_buf);
        }

        pg_res = PQexec(pg_conn, sql_buf);

        if (PQresultStatus(pg_res) != PGRES_TUPLES_OK && PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
            PQclear(pg_res);
            log_postgres_error();
            return;
        }

        PQclear(pg_res);

    } else { // not kill

        if (get_do_last_tables() == 1) {


            snprintf(sql_buf, 4096, "\
                UPDATE aprs_lastobject SET\
                    packet_date = now(),\
                    wind_direction = %d,\
                    wind_speed = %d,\
                    wind_gust = %d,\
                    wind_sustained = %d,\
                    temperature = %d,\
                    rain_hour = %f,\
                    rain_calendar_day = %f,\
                    rain_24hour_day = %f,\
                    humidity = %d,\
                    barometer = %f,\
                    luminosity = %d,\
                    symbol_table = '%s',\
                    symbol_code = '%s',\
                    comment = '%s',\
                    course = %f,\
                    speed = %f,\
                    altitude = %f,\
                    position = GeometryFromText('POINT(%f %f)', -1)\
                WHERE\
                    source = '%s' AND\
                    name = '%s'\
                ", 

                obj->wx.wind_direction,
                obj->wx.wind_speed,
                obj->wx.wind_gust,
                obj->wx.wind_sustained,
                obj->wx.temperature,
                obj->wx.rain_hour,
                obj->wx.rain_calendar_day,
                obj->wx.rain_24hour_day,
                obj->wx.humidity,
                obj->wx.barometer,
                obj->wx.luminosity,
                postgres_escape(&obj->stat.symbol_table, 1, &newlen), 
                postgres_escape(&obj->stat.symbol_code, 1, &newlen),
                postgres_escape(obj->stat.comment, strlen(obj->stat.comment), &newlen), 
                obj->pos.course_made_good, 
                obj->pos.speed_over_ground, 
                obj->pos.altitude, 
                obj->pos.longitude,
                obj->pos.latitude, 
                postgres_escape(obj->source, strlen(obj->source), &newlen),
                postgres_escape(obj->name, strlen(obj->name), &newlen)
                );

            if (get_verbose() & VERBOSE_SQL)
                printf("SQL: %s\n", sql_buf);

            pg_res = PQexec(pg_conn, sql_buf);
            
            if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
                PQclear(pg_res);
                log_postgres_error();
            } else {
                count = atoi(PQcmdTuples(pg_res));
                PQclear(pg_res);
            }

            // If the update affected zero rows, insert the new record.
            if (count == 0) {
                snprintf(sql_buf, 4096, "\
                INSERT INTO aprs_lastobject (\
                    source,\
                    name,\
                    wind_direction,\
                    wind_speed,\
                    wind_gust,\
                    wind_sustained,\
                    temperature,\
                    rain_hour,\
                    rain_calendar_day,\
                    rain_24hour_day,\
                    humidity,\
                    barometer,\
                    luminosity,\
                    symbol_table,\
                    symbol_code,\
                    comment,\
                    course,\
                    speed,\
                    altitude,\
                    position\
                ) VALUES (\
                    '%s',\
                    '%s',\
                    %d,\
                    %d,\
                    %d,\
                    %d,\
                    %d,\
                    %f,\
                    %f,\
                    %f,\
                    %d,\
                    %f,\
                    %d,\
                    '%s',\
                    '%s',\
                    '%s',\
                    %f,\
                    %f,\
                    %f,\
                    GeometryFromText('POINT(%f %f)', -1)\
                )", 

                postgres_escape(obj->source, strlen(obj->source), &newlen),
                postgres_escape(obj->name, strlen(obj->name), &newlen),
                obj->wx.wind_direction,
                obj->wx.wind_speed,
                obj->wx.wind_gust,
                obj->wx.wind_sustained,
                obj->wx.temperature,
                obj->wx.rain_hour,
                obj->wx.rain_calendar_day,
                obj->wx.rain_24hour_day,
                obj->wx.humidity,
                obj->wx.barometer,
                obj->wx.luminosity,
                postgres_escape(&obj->stat.symbol_table, 1, &newlen), 
                postgres_escape(&obj->stat.symbol_code, 1, &newlen),
                postgres_escape(obj->stat.comment, strlen(obj->stat.comment), &newlen), 
                obj->pos.course_made_good, 
                obj->pos.speed_over_ground, 
                obj->pos.altitude, 
                obj->pos.longitude,
                obj->pos.latitude
                );

                if (get_verbose() & VERBOSE_SQL)
                    printf("SQL: %s\n", sql_buf);

                pg_res = PQexec(pg_conn, sql_buf);
                
                if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
                    PQclear(pg_res);
                    log_postgres_error();
                } else {
                    PQclear(pg_res);
                }
            }
        }
    }
}


void item_insert(struct item *itm, unsigned long packet_id) {
	struct sql_insert sql_query;
	int newlen, count;

    snprintf(sql_buf, 4096, "\
        INSERT INTO aprs_item (\
            source,\
            name,\
            symbol_table,\
            symbol_code,\
            comment,\
            position\
        ) VALUES (\
            '%s',\
            '%s',\
            '%s',\
            '%s',\
            '%s',\
            GeometryFromText('POINT(%f %f)', -1)\
        )", 

        postgres_escape(itm->source, strlen(itm->source), &newlen),
        postgres_escape(itm->name, strlen(itm->name), &newlen),
        postgres_escape(&itm->stat.symbol_table, 1, &newlen), 
        postgres_escape(&itm->stat.symbol_code, 1, &newlen),
        postgres_escape(itm->stat.comment, strlen(itm->stat.comment), &newlen), 
        itm->pos.longitude,
        itm->pos.latitude
        );

    if (get_verbose() & VERBOSE_SQL)
        printf("SQL: %s\n", sql_buf);

    pg_res = PQexec(pg_conn, sql_buf);

    if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
        PQclear(pg_res);
        log_postgres_error();
    } else {
        PQclear(pg_res);
    }


    if (itm->kill == 1) {

        snprintf(sql_buf, 4096, "delete from aprs_lastitem where source = '%s' and name = '%s'", postgres_escape(itm->source, strlen(itm->source), &newlen), postgres_escape(itm->name, strlen(itm->name), &newlen));
        pg_res = PQexec(pg_conn, sql_buf);

        if (PQresultStatus(pg_res) != PGRES_TUPLES_OK && PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
            PQclear(pg_res);
            log_postgres_error();
            return;
        }

        PQclear(pg_res);

    } else { // not kill

        if (get_do_last_tables() == 1) {

            snprintf(sql_buf, 4096, "\
                UPDATE aprs_lastitem SET\
                    packet_date = now(),\
                    symbol_table = '%s',\
                    symbol_code = '%s',\
                    comment = '%s',\
                    position = GeometryFromText('POINT(%f %f)', -1)\
                WHERE\
                    source = '%s' AND\
                    name = '%s'\
                ", 

                postgres_escape(&itm->stat.symbol_table, 1, &newlen), 
                postgres_escape(&itm->stat.symbol_code, 1, &newlen),
                postgres_escape(itm->stat.comment, strlen(itm->stat.comment), &newlen), 
                itm->pos.longitude,
                itm->pos.latitude, 
                postgres_escape(itm->source, strlen(itm->source), &newlen),
                postgres_escape(itm->name, strlen(itm->name), &newlen)
                );

            if (get_verbose() & VERBOSE_SQL)
                printf("SQL: %s\n", sql_buf);

            pg_res = PQexec(pg_conn, sql_buf);
                    
            if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
                PQclear(pg_res);
                log_postgres_error();
            } else {
                count = atoi(PQcmdTuples(pg_res));
                PQclear(pg_res);
            }

            // If the update affected zero rows, insert the new record.
            if (count == 0) {
                snprintf(sql_buf, 4096, "\
                INSERT INTO aprs_lastitem (\
                    source,\
                    name,\
                    symbol_table,\
                    symbol_code,\
                    comment,\
                    position\
                ) VALUES (\
                    '%s',\
                    '%s',\
                    '%s',\
                    '%s',\
                    '%s',\
                    GeometryFromText('POINT(%f %f)', -1)\
                )", 

                postgres_escape(itm->source, strlen(itm->source), &newlen),
                postgres_escape(itm->name, strlen(itm->name), &newlen),
                postgres_escape(&itm->stat.symbol_table, 1, &newlen), 
                postgres_escape(&itm->stat.symbol_code, 1, &newlen),
                postgres_escape(itm->stat.comment, strlen(itm->stat.comment), &newlen), 
                itm->pos.longitude,
                itm->pos.latitude
                );

                if (get_verbose() & VERBOSE_SQL)
                    printf("SQL: %s\n", sql_buf);

                pg_res = PQexec(pg_conn, sql_buf);
                
                if (PQresultStatus(pg_res) != PGRES_COMMAND_OK) {
                    PQclear(pg_res);
                    log_postgres_error();
                } else {
                    PQclear(pg_res);
                }
            }
        }
    }  
}
