/*
 * vim:ts=3:

APRS Parser with Database Support
Copyright (C) 2008 James Jefferson Jarvis	<kb0thn@aprsworld.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* I think this motto was in mind when APRS was designed:
 * "Never make anything simple and efficient when a way can be found to make it
 * complex and wonderful" -- Found in the fortune
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <time.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <signal.h>
#include <errno.h>
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_WINSOCK_H
/* In case of insanity break glass. */
#include <winsock.h>
#endif
#include <math.h>
#include <pwd.h>

#include "message.h"
#include "metric.h"
#include "ultimeter.h"


/* variables */
unsigned int insert_mask = INSERT_RAW      |	/* which packets to insert */
									INSERT_MESSAGE  |
									INSERT_POSITION |
/*									INSERT_STATUS   | */
									INSERT_WEATHER  |
									INSERT_LAST     |
									INSERT_WATCH    |
									INSERT_USER     |
									INSERT_TELEM;
/* insert masks are defined in message.h */

unsigned int timeout_reconnect = 0;

unsigned int verbose = VERBOSE_NORM;			// default debugging level
/* verbose masks are defined in message.h */

unsigned int insert_information=1;

/* only insert from one station */
int only_flag=0;
char only_source[10];

int packets_per_commit = 1;
long total_packet_count = 0;
time_t total_packet_count_marker;

int do_last_tables = 1;

/* Internet stuff */
struct sockaddr_in their_addr;	// connector's address information 
int sockfd;

/* function prototypes */
double atof ();

int make_connection (char *host, char *username, char *password, char *database);
int db_begin_txn(void);
int db_commit_txn(void);
int db_rollback_txn(void);

double dm2dd(double dm);
void substr (char *dest, char *src, int size);
char *ltrim(char *s);
char *rtrim(char *s);
int extract_powergain (char *info, struct status *stat);
int extract_altitude (char *info, char *altitude);
void telemetry_coefficients(char *p, char *source);
void telemetry_labels_units(char *p, char *source, char *table);
void telemetry_bits(char *p, char *source);
int parse_telemetry (struct telemetry *telem, char *p);
void telemetry_insert (struct telemetry *telem, unsigned long packet_id);
void user_insert(char *source, char *information, unsigned long packet_id);
int parse_wx (struct weather *wx, char *p);
int parse_compressed_object(struct object *obj, char *p);
int parse_object (struct object *obj, char *p);
int parse_item (struct item *itm, char *p);
void weather_insert (struct weather *wx, unsigned long packet_id);
void object_insert (struct object *obj, unsigned long packet_id);
void item_insert (struct item *itm, unsigned long packet_id);
unsigned char nmea_0183_checksum (char *p);
void extract_gridsquare (struct position *pos, char *square);
void parse_status (struct status *stat, struct position *pos, char *p, unsigned long packet_id);
int extract_comp_position(struct position *pos, struct status *stat, char *info, unsigned long packet_id);
void parse_position (struct position *pos, struct status *stat, char *p, unsigned long packet_id);
int parse_nmea (struct position *pos, char *p);
void lastposition_insert (struct position *pos, struct status *stat, unsigned long packet_id);
void position_insert (struct position *pos, unsigned long packet_id);
void status_insert (struct status *stat, unsigned long packet_id);
void message_insert (struct message *msg, unsigned long packet_id);
unsigned long packet_insert (struct raw_packet *raw);
int parse_tnc (struct raw_packet *raw, char *p);
int parse_message (struct message *msg, char *p);
void parse_mic_e (struct position *pos, struct status *stat, char *p, char *d, unsigned long packet_id);
void inet_feed(char *hostname, int port, char *username, char *password, char *filter);
void stdin_feed(void);
int main (int argc, char *argv[]);
int inet_connect(char *hostname, int port, char *username, char *password, char *filter);
int valid_packet(char *buffer);
void watch_trigger(char *callsign, unsigned long packet_id);
void process (char *buffer);
void init_telemetry(struct telemetry *telem, char *source);
void init_weather(struct weather *wx, char *source);
void init_message(struct message *msg, char *source);
void init_status(struct status *stat, char *source);
void init_position(struct position *pos, char *source);
void init_object(struct object *obj, char *source);
void init_item(struct item *itm, char *source);
int extract_weather_item(char *data, char type, int datalen, char *temp);

/* Yarg, these should be done in a cleaner way. */
unsigned int get_insert_information(void){
	return insert_information;
}
unsigned int get_verbose(void){
	return verbose;
}
unsigned int get_insert_mask(void) {
	return insert_mask;
}

int get_do_last_tables(void){
    /* why isn't this unsigned? */
    return do_last_tables;
}


/* Take degree.minute and return degree.decimal */
double dm2dd(double dm) {
	double dd;
	
	if ( dm > 0 ) {
		dd= floor(dm) + ( (dm - floor(dm)) / 0.6);
	} else {
		dd= ceil(dm) + ( (dm - ceil(dm)) / 0.6);
	}

	return dd;
}

/**
 *  substring function WITH a terminating NULL char, needs a string 
 *  of at least size+1
 */
void substr (char *dest, char *src, int size) {
	strncpy (dest, src, (size_t) size);
	dest[size] = '\0';
}

/* trim functions */
char *ltrim(char *s) {
	char *BeginS = s;

	if (!s || !strlen(s))
		return s;

	while (*s) {
		if (*s != 32) {
			strcpy(BeginS,s);
			break;
		}
		*s++;
	}
	return BeginS;
}

char *rtrim(char *s) {
	char *BeginS = s;
	char *EndS = s + strlen(s) - 1;

	if (!s || !strlen(s))
		return s;


	while (*EndS) {
		if (*EndS != 32) {
			*(EndS + 1) = '\0';
			break;
		}
		*EndS--;
	}
	return BeginS;
}

// APRS Data Extensions               [APRS Reference p.27]
//  .../...  Course & Speed, may be followed by others (see p.27)
//  .../...  Wind Dir and Speed
//  PHG....  Station Power and Effective Antenna Height/Gain
//  RNG....  Pre-Calculated Radio Range
//  DFS....  DF Signal Strength and Effective Antenna Height/Gain
//  T../C..  Area Object Descriptor

/**
 *   Extract powergain from APRS info field          
 *  "PHG1234/" "PHG1234"  from APRS data extension
 * originally from xastir
 *
 * includes the optional RATE specifier which is the rate which the station transmits
 * 
 */
int extract_powergain (char *info, struct status *stat) {
	int i, found, len, phg_len;
	char phgd[10];

	found = 0;
	len = (int) strlen(info);

	if (len >= 7 && strncmp (info, "PHG", 3) == 0 && isdigit ((int) info[3]) && isdigit ((int) info[5]) && isdigit ((int) info[6])) {

		if ( len >= 9 && '/' == info[8] && isalnum((int) info[7]) ) {
			/* contains the rate */
			if ( info[7] >= 'A' && info[7] <= 'Z' ) {
				stat->rate = info[7] - 'A' + 10;	/* 'A' signifies 10 per hour */
			} else if ( info[7] >= '0' && info[7] <= '9' ) {
				stat->rate = info[7] - '0';
			} 
//printf("stat->rate=%d\n",stat->rate);	

			phg_len=9;
		} else {
			phg_len=7;
		}

		
		substr (phgd, info + 3, 4);
		for (i = 0; i <= len - phg_len; i++) {
			// delete powergain from data extension field
			info[i] = info[i + phg_len];
		}

		if (get_verbose() & VERBOSE_DEBUG) {
			// Convert code to actual values per APRS spec and set those values in the status struct
			printf ("phgd[0] = %d\n", phgd[0] - 48);
			printf ("phgd[1] = %d\n", phgd[1] - 48);
			printf ("phgd[2] = %d\n", phgd[2] - 48);
			printf ("phgd[3] = %d\n", phgd[3] - 48);
		}

		/* do the conversions */
		stat->power = pow ((phgd[0] - 48), 2);
		stat->height = FT_TO_M(10 * pow (2, (phgd[1] - 48)));	// spec says this can be a higher ASCII value then
		// a digit - '0' + 25 seems reasonable
		stat->gain = pow (10, (phgd[2] - 48) / 10.0);

		if ( phgd[3] >= '0' && phgd[3] <= '8' ) {
			/* have valid direction */
			stat->directivity = phgd[3] - '0';
		}
	
		if (get_verbose() & VERBOSE_DEBUG) {
			printf ("stat->power = %d\n", stat->power);
			printf ("stat->height = %d\n", stat->height);
			printf ("stat->gain = %f\n", stat->gain);
			printf ("stat->directivity = %d\n", stat->directivity);
		}

	 } else {
		phgd[0] = '\0';
		return (0);
	}
}


/**
 * Taken from XAstir 1.0 source 
 */
int extract_altitude (char *info, char *altitude) {
	int i, ofs, found, len;
	found = 0;

	len = (int) strlen (info);
	for (ofs = 0; (found == 0 && ofs < (len - 8)); ofs++) // search for /A= 
		if (strncmp (info + ofs, "/A=", 3) == 0) {
			found = 1;
			if (!isdigit ((int) info[ofs + 3]) && info[ofs + 3] != '-')
				found = 0;
			for (i = 4; found && i < 9; i++)	// check data format
				if (!isdigit ((int) info[ofs + i]))
			found = 0;
		}
	if (found) {
		ofs--;		// was one too much on exit from for loop
		substr (altitude, info + ofs + 3, 6);
		for (i = ofs; i <= len - 9; i++)	// delete altitude from info field
			info[i] = info[i + 9];
	} else {
		altitude[0] = '\0';
	}
	return (found);
}

void telemetry_bits(char *p, char *source) {
	int i;

	/* stub ... do sometime, but not many people use it */

}


int parse_telemetry (struct telemetry *telem, char *p) {
	/* KB0THN>APRS:T#516,194,031,025,020,117,00010001 */
	int i;


	/* drop out if string is too short */
	if ( strlen(p) < 34 ) {
		if ( get_verbose() & VERBOSE_ERRORS ) {
			printf("Telemetry string is too short\n");
		}
		return 1;
	}

	p += 2;	/* strip of the T# */
	p[3]='\0';

	if ( ! isdigit(p[0]) ) {
		if ( get_verbose() & VERBOSE_ERRORS ) {
			printf("Telemetry sequence number is not a number!\n");
		}
		
		telem->sequence=0;
	} else {
		telem->sequence = atoi(p);
	}
	p += 4;

	/* Read the analog values */
	for ( i=0 ; i<5 ; i++ ) {
		p[3]='\0';
		telem->analog[i] = atoi(p);
		p += 4;
	}

	/* read the digital value */
	telem->digital=0;
	for ( i=0 ; i<8 ; i++ ) {
		if ( '1' == p[i] ) {
			telem->digital += pow(2,7-i);
		}
	}
	p += 8;

	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("telem source: %s\n",telem->source);
		printf("telem sequence: %d\n",telem->sequence);
		for ( i=0 ; i<5 ; i++ ) {
			printf("telem analog[%d]: %d\n",i,telem->analog[i]);
		}
		printf("telem digital: %d\n",telem->digital);
	}
	return 0;
}


/** 
 * Parse weather string and fills the structure if the weather element has data associated with it 
 */
int parse_wx (struct weather *wx, char *p) {
	char state = '_';
	int i=0;
	int flag = 0;
	char *p_end = strlen (p) + p;
	int digits=0;
	int neg_flag=0;

	for (; p < p_end; p++) {
		// printf("\t\t\tstate=%c i=%d digits=%d flag=%d\n",state,i,digits,flag);
		
		/* character at a time atoi */
		if ('-' == *p) {
			neg_flag=1;
			continue;
		}
		if ('0' <= *p && '9' >= *p) {
			i *= 10;
			i += *p - '0';
			digits++;
			continue;
		}

	if ( neg_flag ) {
		i = 0 - i;
	}

	switch (state) {
		case '_':
			if (0 == (flag & 0x1) && digits) {
				wx->wind_direction = i;
				flag |= 0x1;
			}
			break;
		case 'c':
			// BUG: We ignore the flag, because an _ in a time could accidentally trigger that
			if ( digits ) {
				wx->wind_direction = i;
				flag |= 0x1;
			}
			break;
		case '/':
			if (0 == (flag & 0x2) && digits) {
				wx->wind_speed = MILE_TO_KM(i);
				flag |= 0x2;
				}
			break;
		case 'g':
			if (0 == (flag & 0x4) && digits) {
				wx->wind_gust = MILE_TO_KM(i);
				flag |= 0x4;
			}
			break;
		case 't':
			if (0 == (flag & 0x8) && digits) {
				wx->temperature = F_TO_C(i);
				flag |= 0x8;
			}
			break;
		case 'r':
			if (0 == (flag & 0x10) && digits) {
				wx->rain_hour = IN_TO_CM(i / 100.0);
				flag |= 0x10;
			}
			break;
		case 'p':
			if (0 == (flag & 0x20) && digits) {
				wx->rain_calendar_day = IN_TO_CM(i / 100.0);
				flag |= 0x20;
			}
			break;
		case 'P':
			if (0 == (flag & 0x40) && digits) {
				wx->rain_24hour_day = IN_TO_CM(i / 100.0);
				flag |= 0x40;
			}
			break;
		case 'h':
			if (0 == (flag & 0x80) && digits) {
				wx->humidity = i;
				flag |= 0x80;
			}
			break;
		case 'b':
			if (0 == (flag & 0x100) && digits) {
				wx->barometer = i / 10.0;
				flag |= 0x100;
			}
			break;
		case 's':	// BUG Also possibly snowfall
			if (0 == (flag & 0x200) && digits) {
				wx->wind_sustained = MILE_TO_KM(i);
				if ( wx->wind_speed == 0 )
					wx->wind_speed = wx->wind_sustained;
				flag |= 0x200;
			}
			break;
		case 'l':
			if (0 == (flag & 0x400) && digits) {
				wx->luminosity = i;
				flag |= 0x400;
			}
			break;
		case 'L':
			if (0 == (flag & 0x400) && digits) {
				wx->luminosity = i + 1000;
				flag |= 0x400;
			}
			break;
		case '#':
			if (0 == (flag & 0x800) && digits) {
				wx->rain_count = i;
				flag |= 0x800;
			}
			break;
	}
	i = 0;
	digits=0;
	
	state = *p;
	}
#if 0
	printf("weather source: %s\n",wx->source);
	printf ("wind_direction: %d \n", wx->wind_direction);
	printf ("temperature: %d \n", wx->temperature);
	printf ("wind_speed: %d \n", wx->wind_speed);
	printf ("wind_gust: %d \n", wx->wind_gust);
	printf ("rain_calendar_day: %f \n", wx->rain_calendar_day);
	printf ("rain_24hour_day: %f \n", wx->rain_24hour_day);
	printf ("rain_hour: %f \n", wx->rain_hour);
	printf ("humidity: %d \n", wx->humidity);
	printf ("barometer: %d \n", wx->barometer);
	printf ("wind_sustained: %d \n", wx->wind_sustained);
	printf ("luminosity: %d \n", wx->luminosity);
	printf ("rain_count: %d \n", wx->rain_count);
#endif
	return flag;
}

/**
 * Calculate NMEA checksum using XOR */
unsigned char nmea_0183_checksum (char *p) {
	int checksum, start, done;
	checksum = start = done = 0;

	for (; *p != '\0'; p++) {
		if (*p == '*')
			done = 1;
		if (start == 1 && done == 0) {
			checksum = checksum ^ *p;	// XOR 
		}
		if (*p == '$')
			start = 1;
	}
	return checksum;
}

/**
 * Fill *pos with the latitude and longitude from an all UPPERCASE grid square 
 */
void extract_gridsquare (struct position *pos, char *square) {
	int i;
	int len = strlen (square);

	if ( get_verbose() & VERBOSE_DEBUG )
		printf("Extracting grid square\n");
	if (square[0] >= 'A' && square[0] <= 'I') {
		pos->longitude = (square[0] - 'A') * 20 - 180;
	} else {
		pos->longitude = (square[0] - 'J') * 20;
	}

	if (square[1] >= 'A' && square[1] <= 'I') {
		pos->latitude = (square[1] - 'A') * 10 - 90;
	} else {
		pos->latitude = (square[1] - 'J') * 10;
	}

	pos->latitude += (square[3] - '0');
	pos->longitude += (square[2] - '0') * 2;

	if (isupper (square[4]) && isupper (square[5])) {
#define LAT	0.0416666
#define LAT_HALF	0.0208333
#define LON	0.0833333
#define LON_HALF	0.0416666
		pos->latitude += (square[5] - 'A') * LAT + LAT_HALF;
		pos->longitude += (square[4] - 'A') * LON + LON_HALF;

		for (i = 0; i <= len - 7; i++)	// Swallow grid square and space
			square[i] = square[i + 7];
	} else {
		pos->latitude += 0.5;	
		pos->longitude += 1;

		for (i = 0; i <= len - 5; i++)	// Swallow grid square and space
			square[i] = square[i + 5];
	}
}

void parse_status (struct status *stat, struct position *pos, char *p, unsigned long packet_id) {
/* Extensions (just to keep life interesting):
 *
 * Maidenhead grid square:
 * "IO91SX/G Hello World"	(PIC-E or MIC-E)
 * ">IO91SX/G Hello World"	(Kenwood D-7)
 * "]IO91SX/G Hello World"	(Kenwood D-700)
 * "The locator must be the first word in the Status Report or status text." - APRS Spec
 */
	char buffer[256];
	char *b;
	int pos_insert = 0;
	struct weather wx;
	int altitude;

	init_weather(&wx,stat->source);

	buffer[0] = '\0';

	if (p[0] == '>')	{ // This is a plain 'ol status packet, not extensions
		 p++;		// swallow the first character before moving on;

		/* Check if this status packet contains a gridsquare */
		if (isupper (p[0]) && isupper (p[1]) && isdigit (p[2]) && isdigit (p[3]))
			if ((isupper (p[4]) && isupper (p[5])) || isspace (p[4]) || p[4] == '\0') {
				extract_gridsquare (pos, p);
				pos_insert++;
			}
	}

	// Check if we have the  123/987	course/speed information
	if (  isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2]) && '/' == p[3] && 
	isdigit(p[4]) && isdigit(p[5]) && isdigit(p[6]) ) {
		memcpy(buffer,p,3);
		buffer[3] = '\0';
		pos->course_made_good = (float) atoi(buffer);
		memcpy(buffer,p+4,3);
		buffer[3] = '\0';
		pos->speed_over_ground = NM_TO_KM((float) atoi(buffer));
		p +=7;	// Swallow this
	}

	if (get_verbose() & VERBOSE_DEBUG)
		printf ("Parse Status started with: %s\n", p);

	extract_powergain (p, stat);
	
	if (0 != extract_altitude (p, buffer)) {
		pos->altitude = FT_TO_M(atof (buffer));
		pos_insert++;
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("Altiude = %f\n", pos->altitude);
	}

	/* check if it a decoded mic-e packet, then clean the garbage out */
	if ( 0 == strncmp(p,"/Mic-E/",7) || 0 == strncmp(p,"/E>dsy/",7) ) {

		p += 10;	// Swallow /Mic-E/Mx/

		strncpy(buffer,p,10);
		buffer[10] = '\0';

		if ( strlen(p) < strlen(buffer) ) {
			// printf("big problems ... this shouldn't have happened\n");
			goto out;	// highly structured code right here! BUG BUG BUG
		}
		p += strlen(buffer);
			
		b = buffer;
		/* get rid of pad characters like ... */
		while ( '\0' != *b && (isalnum(*b) || isspace(*b)) ) {
			b++;
		}
		*b='\0';


		if ( get_verbose() & VERBOSE_DEBUG ) {
			printf("--> Mic-E packet <--\n");
			printf("\tMic-E status message: %s\n",buffer);
			printf("\tstrlen(p)=%d strlen(buffer)=%d\n",strlen(p),strlen(buffer));
		}


		if ( strlen(p) ) {
			p++;

			/* Check for Kenwood D7 or D700 */
			if ( '>' == *p ) {
				if ( get_verbose() & VERBOSE_DEBUG )
					printf("\tKenwood D7\n");
				strcat(buffer," Kenwood D7");
				p++;
			} else if ( ']' == *p ) {
				if ( get_verbose() & VERBOSE_DEBUG )
					printf("\tKenwood D700\n");
				strcat(buffer," Kenwood D700");
				p++;
			}

			/* P may now point to altitude information */
			if ( strlen(p) > 3 && '}' == p[3] ) {

				altitude = (*p - 33) * 8281;	// 91^2
				p++;
				altitude += (*p - 33) * 91;
				p++;
				altitude += (*p -33);
				p++;
	
				altitude -= 10000;	// referenced to 10,000 meters below sea level

				pos->altitude=altitude;

				if ( get_verbose() & VERBOSE_DEBUG ) {
					printf("\tAltitude information found\n");
					printf("\tAltitude = %d meters   %d feet\n",pos->altitude,(int) M_TO_FT( pos->altitude));
				}
	
				p++;	// swallow the 
			}
		}

out:
		// left trim
		while ( '\0' != *p && isspace(*p) )
			p++;

		/* combine the clean MIC-E status message with the user input text */
		if ( strlen(p) > 0 ) 
			strcat(buffer,": ");
		strncat(buffer,p,256);
		p=buffer;

	}


	/* left trim */
	while ( '\0' != *p && isspace(*p) )
		p++;

	if (get_verbose() & VERBOSE_DEBUG) {
		printf ("Parse Status ended with: %s\n", p);
	}
	strncpy (stat->comment, p, 256);	// populate stat->comment with what is left

	pos->time_of_fix = 0;	// let the raw table handle the time


	// Weather stations could be "hidden" ... we have to look at its symbol to find out
	// if it is a weather sation. If it is a weather station then we don't want to have it's
	// position have a course and speed (because those are weather readings, not position readings)
	if ( (stat->symbol_table == '\\' || stat->symbol_table == '/') && stat->symbol_code == '_' ) {
		if ( get_verbose() & VERBOSE_DEBUG)
			printf("###########This appears to be a weather station.######################\n");
		if ( parse_wx (&wx, stat->comment) ) {
			if ( -1 == wx.wind_direction )
				wx.wind_direction = pos->course_made_good;
			if ( -1 == wx.wind_speed )
				wx.wind_speed = pos->speed_over_ground;
			pos->course_made_good=-1;	// reset to NULL values
			pos->speed_over_ground=-1;
		
			weather_insert (&wx, packet_id);
		}
		packet_count[WEATHER]++;
	}

	status_insert (stat, packet_id);
}

int extract_comp_position(struct position *pos, struct status *stat, char *info, unsigned long packet_id) {
	int i=0;

	char *p=info;
	int range=0;
	
//	strcpy(info,"=/6msc7TbEx Jim on Madeline Island");
//	strcpy(info,"=/5L!!<*e7>7P[ with speed=32 knots and course=88 degrees");
//	strcpy(info,"=/5L!!<*e7>S]\024 altitude 10004 feet");
//	strcpy(info,"=/97Uu5pBD- BWelcome to Norfolk, Enjoy");
	p++;

	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("in extract_comp_position (raw is %s)\n",info);
	}

	/* verify to see if our characters are in range */
	for ( i=1 ; i<9 ; i++ ) {
		if ( 9 == i || 12 == i )
			continue;

		if ( p[i] < '!' || p[i] > '{' ) {
//			printf("Compressed report contains invalid character at %d\n",i);
			return -1;
		}
		
	}
	stat->symbol_table=p[0];
	stat->symbol_code=p[9];
	strncpy(stat->comment,&info[12],256);

	pos->latitude = 90 - ( (p[1]-'!')*753571 + (p[2]-'!')*8281 + (p[3]-'!')*91 + (p[4]-'!') ) / 380926.0;
	pos->longitude = -180 + ( (p[5]-'!')*753571 + (p[6]-'!')*8281 + (p[7]-'!')*91 + (p[8]-'!') ) / 190463.0;
	
	if ( ' ' != p[10] ) {
		if ( 16==(p[12]&24) ) {
			/* contains altitude information because bits 4,3 = 1,0 */
			pos->altitude = (int) pow(1.002,( ((p[10]-'!')*91) + p[11]-'!'));
		} else if ( p[10] >= '!' && p[10] <= 'z' ) {
			/* course in degrees */
			pos->course_made_good = (float) (p[10]-'!')*4;
			/* speed in knots */
			pos->speed_over_ground = (float) NM_TO_KM(pow(1.08,(p[11]-'!')) - 1);
		} else if ( '{' == p[10] ) {
			//printf("radio range");
			/* pre-calculated radio range */
			range = (int) (2.0 * pow(1.09,(p[11]-'!')));
		}
	}
	
	/* bug: put the range into the stat->comment field */
	
	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("latitude=%lf\n",pos->latitude);
		printf("longitude=%lf\n",pos->longitude);
		printf("altitude=%d\n",pos->altitude);
		printf("symbol_table=%c\n",stat->symbol_table);
		printf("symbol_code=%c\n",stat->symbol_code);
		printf("course=%d\n",pos->course_made_good);
		printf("speed=%f\n",pos->speed_over_ground);
		printf("pre-calculated radio range=%d\n",range);
	}
	

//	printf("--> all done, we are left with: %s\n",info);
	
	/* anything left is status information */
	status_insert (stat, packet_id);

	
	return 0;
}

void parse_position (struct position *pos, struct status *stat, char *p, unsigned long packet_id) {
	char buffer[256];

  
	if ('/' == *p || '@' == *p) {
		/* eat the time, will look like this @252054z */
		p += 7;
	}

	/* compressed reports will start with non-digit */
	if ( ! isdigit(*(p+1)) &&  0 == extract_comp_position(pos, stat, p, packet_id) ) {
		/* we are done, compressed portion contains everything */
		return;
	} 

	p++;		// skip first character


	/* latitude */
	strncpy (buffer, p, 7);	
	buffer[7] = '\0';

	pos->latitude = dm2dd(atof (buffer) / 100);
	p += 7;
	if ('S' == *p)
		pos->latitude = 0 - pos->latitude;
	p++;

	/* symbol table */
	stat->symbol_table = *p;	
	p++;

	/* longitude */
	strncpy (buffer, p, 8);	// Longitude
	buffer[8] = '\0';
	pos->longitude = dm2dd(atof (buffer) / 100);
	p += 8;
	if ('W' == *p)
		pos->longitude = 0 - pos->longitude;
	p++;

	/* symbol code */
	stat->symbol_code = *p;
	p++;

	parse_status (stat, pos, p, packet_id);


	if (get_verbose() & VERBOSE_DEBUG) {
		printf ("latitude: %f\n", pos->latitude);
		printf ("symbol table: %c\n", stat->symbol_table);
		printf ("longitude: %f\n", pos->longitude);
		printf ("symbol code: %c\n", stat->symbol_code);
		printf ("comment: %s\n", stat->comment);
	}

	// BUG: Use packet time if available, or use system time

}

int parse_nmea (struct position *pos, char *p) {
	char *p_end = strlen (p) + p;
	char elements[20][20];	// Store each element until we can put in structure
	int i, element, last, checksum;
	unsigned char local_checksum = nmea_0183_checksum (p);
	struct tm tomake;
	struct tm *curtime;
	time_t aprstime;

	time (&aprstime);
	curtime = gmtime (&aprstime);


	/* First parse to break into elements, then let the sentence parsers have fun */

	// initialize
	for (i = 0; i < 20; i++)
		strcpy (elements[i], "");

	i = 0;
	element = 0;
	for ( ; (element < 20) && (p < p_end); p++, i++) {
		if (last == ',')
			i = 0;
      if (*p == ',') {
			elements[element][i] = '\0';	// make it a string!    
			i = 0;
			element++;
		}
		elements[element][i] = last = *p;
	}
	elements[element][i] = '\0';  // make it a string!    

	if (get_verbose() & VERBOSE_DEBUG) {
		for (i = 0; (i < 20); i++) {
			if (strcmp (elements[i], ""))
				printf ("elements[%d]: %s\n", i, elements[i]);
		}
	}
	if (strstr (elements[0], "PNTS")) {
		if (get_verbose() & VERBOSE_PARSE)
			fprintf (stdout, "NOT IMPLEMENTED: NAVI-TRA Packets\n");
		return -1;
	} else if (strstr (elements[0], "RMC")) {
// $GPRMC,222128.00,A,4211.1283,N,08805.4656,W,0.0,341.5,160301,,*24
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("GPRMC sentence\n");
		strncpy (pos->sentence, elements[0], 6);
		// Time
		p = elements[1];

		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_hour = (atoi (elements[19]));
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_min = (atoi (elements[19]));
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_sec = (atoi (elements[19]));

		pos->rx_warning = elements[2][0];

		pos->latitude = dm2dd(atof (elements[3]) / 100);
		if (elements[4][0] == 'S')
			pos->latitude = (0 - pos->latitude);	// make negative

		pos->longitude = dm2dd(atof (elements[5]) / 100);
		if (elements[6][0] == 'W')
			pos->longitude = (0 - pos->longitude);	// make negative

		pos->speed_over_ground = NM_TO_KM(atof (elements[7]));
		pos->course_made_good = atof (elements[8]);
		// date
		p = elements[9];
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_mday = (atoi (elements[19]));
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_mon = (atoi (elements[19]) - 1);
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_year = (100 + atoi (elements[19]));
		if (get_verbose() & VERBOSE_DEBUG) {
			printf ("day: %d\n", tomake.tm_mday);
			printf ("month: %d\n", tomake.tm_mon);
			printf ("year: %d\n", tomake.tm_year);
		}

		aprstime = mktime (&tomake);
		pos->time_of_fix = aprstime;

		if (get_verbose() & VERBOSE_DEBUG)
			printf ("%lu\n", aprstime);

		pos->magnetic_variance = atof (elements[10]);

		/* Spec says that checksum should be in 11 ... but KB7RSI-15 breaks
		 * that and puts it in 12 ... and aprsdec still says okay... ???
		 * * $GPRMC,222002,A,3614.4133,N,11509.6897,W,0.0,355.3,070401,13.1,E,A*31
		 */
		if (strcmp (elements[12], "")) {
			p = elements[12];
			p_end = strlen (elements[12]) + p;
		} else {
			p = elements[11];
			p_end = strlen (elements[11]) + p;
		}
		for (; p < p_end && *p != '*'; p++)
			;
		p++;		// move to beginning of hex

		checksum = 0;
		sscanf (p, "%2x", &checksum);
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("Scanned checksum: %02x Calculated: %02x\n", checksum,local_checksum);

		if (checksum == local_checksum) {
			pos->checksum = 1;
			if (get_verbose() & VERBOSE_DEBUG)
				printf ("good checksum\n");
		} else {
			pos->checksum = 0;
			if (get_verbose() & VERBOSE_PARSE)
				printf ("bad checksum\n");
			return -1;
		}
	} else if (strstr (elements[0], "GGA")) {
		/* $GPGGA,223623,3601.917,N,07952.321,W,1,03,2.1,291.7,M,-32.7,M,,*70 */
		//  0      1      2       3  4        5 6 7  8   9     0  11   2   14
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("GPGGA sentence\n");
		strncpy (pos->sentence, elements[0], 6);
		// time
		p = elements[1];

		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_hour = (atoi (elements[19]));
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_min = (atoi (elements[19]));
		elements[19][0] = *p++;
		elements[19][1] = *p++;
		elements[19][2] = '\0';
		tomake.tm_sec = (atoi (elements[19]));
		
		tomake.tm_mday = curtime->tm_mday;
		tomake.tm_mon = curtime->tm_mon;
		tomake.tm_year = curtime->tm_year;
		
		pos->time_of_fix = aprstime = mktime (&tomake);

		pos->latitude = dm2dd(atof (elements[2]) / 100);
		if (elements[3][0] == 'S')
			pos->latitude = (0 - pos->latitude);	// make negative

		pos->longitude = dm2dd(atof (elements[4]) / 100);
		if (elements[5][0] == 'W')
			pos->longitude = (0 - pos->longitude);	// make negative

		pos->quality = atoi (elements[6]);
		pos->sats = atoi (elements[7]);
		pos->hor_dilution = atof (elements[8]);
		pos->altitude = atof (elements[9]);
		// BUG: parse units - is it ever anything besides meters???
		pos->geoidal_separation = atof (elements[11]);
      // BUG: parse units - ditto
		pos->diff_last_update = atof (elements[13]);

		p = elements[14];
		if ('*' == *p) {
			// Just read checksum
			sscanf (p, "%*c%2x", &checksum);
		} else {
			// Diffrential station ID and checksum
			sscanf (p, "%4d%*c%2x", &pos->diff_station, &checksum);
		}
		if (checksum == local_checksum) {
			pos->checksum = 1;
			if (get_verbose() & VERBOSE_DEBUG)
				printf ("good checksum\n");
		} else {
		pos->checksum = 0;
		if (get_verbose() & VERBOSE_PARSE)
			printf ("bad checksum\n");
		return -1;
		}
		
		if (get_verbose() & VERBOSE_DEBUG) {
			printf ("Date: %ld\n", pos->time_of_fix);
			printf ("Latitude: %f\n", pos->latitude);
			printf ("Longitude: %f\n", pos->longitude);
			printf ("Quality: %d\n", pos->quality);
			printf ("Satellites in use: %d\n", pos->sats);
			printf ("Horizontal Dilution: %f\n", pos->hor_dilution);
			printf ("Altitude: %f\n", pos->altitude);
			printf ("Geoidal Separation: %f\n", pos->geoidal_separation);
			printf ("Last DGPS update: %f\n", pos->diff_last_update);
		}
	} else if (strstr (elements[0], "GLL")) {
		if (get_verbose() & VERBOSE_PARSE)
			fprintf (stdout, "NOT IMPLEMENTED: NMEA GLL\n");
		return -1;
	} else if (strstr (elements[0], "VTG")) {
		if (get_verbose() & VERBOSE_PARSE)
			fprintf (stdout, "NOT IMPLEMENTED: NMEA VTG\n");
		return -1;
	} else if (strstr (elements[0], "WPT")) {
		if (get_verbose() & VERBOSE_PARSE)
			fprintf (stdout, "NOT IMPLEMENTED: NMEA WPT\n");
		return -1;
	}
return 0;	// successful, otherwise we would have already returned
}


/**
 * parse_tnc converts TNC address information (per spec) into struct raw_packet values 
 */
int parse_tnc (struct raw_packet *raw, char *p) {
	/* APRS spec (page 82) describes the TNC2 and AEA formats:
	TNC2:	source>dest[,dest,...]*:information
	AEA:	source>dest[>dest>...]*:information
	* (asterick) indicates the digipeater the packet was heard from

	Captured from second.aprs.net (port 23):
	TNC2:	KB2WYA>BEACON,WA2DQL-1*,WIDE,WIDE:
	AEA:	KB7RSI-15>N6BG-4*>WIDE>GPS:
	*/
	char *p_end = strlen (p) + p;
	char temp[128];	// use until we decide if destination or digipeater
	int i, j, digi;
	int next, previous;	// keep track of what our next and previous delimiters
	// so we can figure destination and digipeaters
	int length = strlen (p);

	strcpy (raw->source, "");	// make sure old stuff didn't stick around
	strcpy (raw->destination, "");

	/* some tncs will show "cmd:" when switching out of command mode */
	if ( p[0] == 'c' && p[1] == 'm' && p[2] == 'd' && p[3] == ':' ) {
		p+=4;
	}

	// Capture source
	for (i = 0; (i < 9) && (p < p_end) && (*p != '>'); p++, i++) {
		if ( ! isalnum(*p) && '-' != *p ) {
			return 1;
		}
		raw->source[i] = *p;
	}

	if ( raw->source[i-1] == '*' ) {
		raw->source[i--];
	}
	raw->source[i] = '\0';	// terminate string

	p++;		// swallow '>'

/* 
TRUTH TABLE:
first field after source:
if ( next non-alpha-numeric-'-' - '*' ) == ':' ) then destination 
if ( next non-alpha-numeric-'-' - '*' ) == ',' ) then destination 
if ( next non-alpha-numeric-'-' - '*' ) == '>' ) then digipeater 
for ( fields after first ):
if ( next non-alpha-numeric-'-' - '*' ) == ':' && 
	( previous non-alpha-number-'-' - '*' ) == '>' ) then destination 
if ( next non-alpha-numeric-'-' - '*' ) == ':' && 
	( previous non-alpha-number-'-' - '*' ) == ',' ) then digipeater 
if ( next non-alpha-numeric-'-' - '*' ) == ',' ) then digipeater
if ( next non-alpha-numeric-'-' - '*' ) == '>' ) then digipeater
*/

	previous = next = '>';
	digi = 0;
	for (i = 0, j = 0; (p < p_end); p++, j++) {
		if (*p == ':' || *p == ',' || *p == '>') {
			previous = next;
			next = *p;
			temp[j++] = '\0';	// terminate string
			/* store the callsign */
			if (i == 0 && (next == ':' || next == ','))
				strncpy (raw->destination, temp, 10);
			else if (i == 0 && next == '>')
				strncpy (raw->digipeater[digi++], temp, 10);
			else if (i > 0 && (next == ':' && previous == '>'))
				strncpy (raw->destination, temp, 10);
			else if (i > 0 && (next == ':' && previous == ','))
				strncpy (raw->digipeater[digi++], temp, 10);
			else if (i > 0 && (next == ',' || next == '>'))
				strncpy (raw->digipeater[digi++], temp, 10);

//			if (get_verbose() & 255)
//				printf ("a callsign: %s next = %c previous = %c i = %d\n", temp, next, previous, i);

			p++;		// swallow the next character
			i++;		// another callsign
			j = 0;	// we have found a delimiter
		
			if (':' == next)
				break;	// all the further we need to go
		}
		temp[j] = *p;
	}

	/* This is one of the new (undefined) spec for identifying who injected the packet
	 * into the internet */
	if ( 0 ==  strcmp(raw->digipeater[digi-1],"I") ) {
		strncpy(raw->ip_source,raw->digipeater[digi-2],10);
		digi -= 2;
	} else {
		strcpy(raw->ip_source,"");	// BUG we should initialize structure instead
	}
	
	for (i = 8; i >= digi && i >= 0; i--)
		strcpy (raw->digipeater[i], "");	// initialize rest of array --no garbage

	// information field
	for (i = 0; (i < 513) && (p < p_end) && (*p != '\n'); p++, i++) {
		raw->information[i] = *p;
	}
	raw->information[i] = '\0';	// terminate string


#if 0
	if (get_verbose() & VERBOSE_DEBUG) {
		printf("IP Source = %s\n",raw->digipeater[digi-2]);
		printf ("source: %s\n", raw->source);
		printf ("destination: %s\n", raw->destination);
		printf ("digipeter[0]: %s\n", raw->digipeater[0]);
		printf ("digipeter[1]: %s\n", raw->digipeater[1]);
		printf ("digipeter[2]: %s\n", raw->digipeater[2]);
		printf ("digipeter[3]: %s\n", raw->digipeater[3]);
		printf ("digipeter[4]: %s\n", raw->digipeater[4]);
		printf ("digipeter[5]: %s\n", raw->digipeater[5]);
		printf ("digipeter[6]: %s\n", raw->digipeater[6]);
		printf ("digipeter[7]: %s\n", raw->digipeater[7]);
		printf ("information: %s\n", raw->information);
	}
#endif
	
	return 0;
}

/**
 * parse_message splits an APRS message into addressee, text, and ID ... if
 * the packet is not really an APRS message it won't crash, but instead split
 * the message at the boundaries and loses a few characters (0,10,68)
*/
int parse_message (struct message *msg, char *p) {
	char *p_end = strlen (p) + p;
	int i;
	int length = strlen (p);

	p++;		// swallow ':'
	for (i = 0; (i < 9) && (p < p_end) && (*p != ':'); p++, i++) {
		msg->addressee[i] = *p;
	}
	msg->addressee[i] = '\0';	// terminate string

	p++;		// swallow ':'
	for (i = 0; (i < 67) && (p < p_end) && (*p != '{'); p++, i++) {
		msg->text[i] = *p;
	}
	msg->text[i] = '\0';

	/* check for telemetry parameters */
	if ( strstr(msg->text,"PARM.") ) {
		telemetry_labels_units(msg->text,msg->addressee,"telemetry_labels");
		return 1;
	} else if ( strstr(msg->text,"UNIT.") ) {
		telemetry_labels_units(msg->text,msg->addressee,"telemetry_units");
		return 2;
	} else if ( strstr(msg->text,"EQNS.") ) {
		telemetry_coefficients(msg->text,msg->addressee);
		return 3;
	} else if ( strstr(msg->text,"BITS.") ) {
		telemetry_bits(msg->text,msg->addressee);
		return 4;
	}

	p++;		// swallow '{'
	for (i = 0; (i < 5) && (p < p_end); p++, i++) {
		msg->id[i] = *p;
	}
	msg->id[i] = '\0';

	if (get_verbose() & VERBOSE_DEBUG) {
		printf ("\tAddressee: %s\n", msg->addressee);
		printf ("\tText: %s\n", msg->text);
		printf ("\tID: %s\n", msg->id);
	}
	return 0;
}

/**
 * Parse MICE packets using APRSd's MIC_E parser (modified)
 */
void parse_mic_e (struct position *pos, struct status *stat, char *p, char *d, unsigned long packet_id) {
	char packet1[256];
	char packet2[256];
	int packet1_len;
	int packet2_len;
	int len;

	len = strlen (p);

	if (0 == fmt_mic_e (d, p, len, packet1, &packet1_len, packet2, &packet2_len)) {
		if (get_verbose() & VERBOSE_PARSE)
			printf ("%snot a valid MIC-E?%s\n", RED_TEXT, DEFAULT_TEXT);
	} else {
		packet1[packet1_len] = '\0';
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("MIC-E has been converted to: %s\n", packet1);

		packet2[packet2_len] = '\0';
		if (get_verbose() & VERBOSE_DEBUG)
			printf ("buff4 = %s\n", packet2);
      
		parse_position (pos, stat, packet1, packet_id);
		position_insert (pos, packet_id);
	}
}


int inet_connect(char *host, int port, char *username, char *password, char *filter) {
	struct hostent *he;
	char loginBuff[256];
	int nSentBytes, nReadBytes;

	if ( get_verbose() & VERBOSE_DEBUG)
    		printf("Connecting to: %s(%d)\n",host,port);	
    	
	if ( (he = gethostbyname (host) ) == NULL ) {
		herror("DNS lookup failed");

		return 1;
	}
	    
	
	if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
    		perror ("socket");

		return 2;
	}

	/*
	if (setsockopt(listen_sd, SOL_SOCKET,  SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
		perror ("socket SO_REUSEADDR");
		return 2;
	}

	if (ioctl(listen_sd, FIONBIO, (char *)&on) < 0) {
		perror ("socket FIONBIO");
		return 2;
	}
	*/

	their_addr.sin_family = AF_INET;	// host byte order
	their_addr.sin_port = htons (port);	// short, network byte order
	memcpy(( caddr_t )&(their_addr.sin_addr), he->h_addr, he->h_length);
	memset(&(their_addr.sin_zero), 0, 8);	// zero the rest of the struct 
       

	if (connect(sockfd, (struct sockaddr *) &their_addr,sizeof (struct sockaddr)) == -1) {
    		perror ("Error connecting");

		return 3;
	}

	/* connected */

	/* send our login string if we have it */
	if ( strlen(username) > 0 && strlen(password) > 0 ) {
		if (strlen(filter) > 0) {
			sprintf(loginBuff,"user %s pass %s vers aprsworld 1.0 filter %s\r\n",username,password,filter);
		} else {
			sprintf(loginBuff,"user %s pass %s\n",username,password);
		}

		if ( get_verbose() & VERBOSE_DEBUG ) 
			printf("Sending login: %s",loginBuff);

		nReadBytes = strlen(loginBuff);
		nSentBytes=0;

		while (nSentBytes < nReadBytes) {
	 		int nTemp = send(sockfd, loginBuff + nSentBytes, nReadBytes - nSentBytes, 0);
			
			if (nTemp > 0) {
  				nSentBytes += nTemp;
			} else {
				fprintf(stderr,"Connection unexpectedly dropped while sending login.\n");
				/* unexpectedly dropped connection while sending login */
				return 4;
			}
		}
	}
		
	/* Set the host id for raw packets  */
	qa_hostid = set_qa_hostid( host );

	return 0;
}

void inet_feed(char *hostname, int port, char *username, char *password, char *filter) {
	char buffer[BUFFSIZE], *p;
	int numbytes;
	int r, res;
	struct timeval timeout;
	fd_set read_set;

	/* begin SQL transaction */
    	db_begin_txn();

	for(;;) {
		r=inet_connect(hostname, port, username, password, filter);
		if ( r ) {
			/* error connecting */
			return;
		}

		/* select/sleep cycle to detect flow interruptions */
		timeout.tv_sec = timeout_reconnect;
		timeout.tv_usec = 0;

		FD_ZERO(&read_set);
		FD_SET(sockfd, &read_set);

		for(;;) {
			res = select(sockfd + 1, &read_set, NULL, NULL, &timeout);

			if (res == -1) {
				perror("select failed");
			} else if (res == 0) {
				if ( get_verbose() & VERBOSE_DEBUG ) 
					printf("Socket timeout.");
				break;
				
			} else {

				numbytes = recv (sockfd, buffer, BUFFSIZE, 0);
				buffer[numbytes] = '\0';
			
				process(strtok(buffer,"\r\n"));
				while ( p = strtok(NULL,"\r\n") ) {
					process(p);
				}
			}
		}

		if (numbytes < 0)	{	// there was an error
			openlog("APRS Parser",LOG_PERROR,LOG_USER);
			syslog(LOG_NOTICE,"Connection closed ... here's why: %s\n",strerror(errno));
			closelog();
		}

	}
}

void stdin_feed(void) {
	char buffer[BUFFSIZE];

	while (0 != fgets (buffer, BUFFSIZE, stdin)) {
		process (buffer);
	}
}

int main (int argc, char *argv[]) {
	char db_database[256], db_host[256], db_user[256], db_pass[256]; /* database information */
	char aUsername[16], aPassword[16]; /* APRS host username and password */
	char filter[256]; /* Connection filter */
	int n, daemon = 0;

	char server_list[256];

	/* Default database values */
	strcpy(db_database, "aprs");	
	strcpy(db_host, "localhost");
	strcpy(db_user, "root");
	strcpy(db_pass, "");
	strcpy(aUsername, "NOCALL");
	strcpy(aPassword,"12960");
	strcpy(filter,"");

	while ((n = getopt (argc, argv, "o:it:lb:h:u:p:dv:c:U:P:f:m:")) != -1) {
		switch (n) {
			case 'm':
				timeout_reconnect = atoi(optarg);
				break;
			case 'f':
				strncpy(filter,optarg,256);
				break;
			case 'U':
				strncpy(aUsername,optarg,16);
				aUsername[15]='\0';
				break;
			case 'P':
				strncpy(aPassword,optarg,16);
				aPassword[15]='\0';
				break;

			/* only process packets with this source callsign */
			case 'o':
				only_flag=1;
				strncpy(only_source,optarg,10);

				if ( get_verbose() & VERBOSE_DEBUG)
					fprintf(stderr,"Only using packets from: %s\n",only_source);
				
				break;
			/* insert information part of packet into raw */
			case 'i':
				insert_information=0;
				break;
			/* Insert masks is a bit-masked integer that controls what packets are inserted into
			 * the database 
			 */
			case 't':
				insert_mask = atoi(optarg);
				if (get_verbose() & VERBOSE_DEBUG)
					fprintf (stderr, "INSERT MASK=0x%02X\n", insert_mask);
				break;

			case 'l':
				printf("Not logging to aprs_last* tables.\n");
				do_last_tables = 0;
				break;

			/* database connection information */
			case 'b':
				strncpy (db_database, optarg, 256);
				break;
			case 'h':
				strncpy (db_host, optarg, 256);
				break;
			case 'u':
				strncpy (db_user, optarg, 256);
				break;
			case 'p':
				strncpy (db_pass, optarg, 256);
				break;

			/* Fork ourselves and become a daemon */
			case 'd':
				daemon = 1;
				if (get_verbose() & VERBOSE_DEBUG)
					fprintf (stderr, "%sStarting as a daemon...%s\n", GREEN_TEXT, DEFAULT_TEXT);
				break;
			case 'v':
				verbose = atoi (optarg);
				if (get_verbose() & VERBOSE_DEBUG)
					fprintf(stderr,"verbose level = 0x%02X\n", get_verbose());
				break;
			case 'c':
				packets_per_commit = atoi (optarg);
				if (get_verbose() & VERBOSE_DEBUG)
					fprintf(stderr,"Commiting for each %d packets.\n", packets_per_commit);
				break;
		}
	}

	if (daemon == 1) {
		if ( ! daemon_start (0) ) {
			fprintf (stderr, "%s: cannot become a daemon\n", argv[0]);
			return 1;
		}
	}
    
	if (make_connection(db_host, db_user, db_pass, db_database)) {
		fprintf (stderr, "%sCould not connect to database%s\n",RED_TEXT, DEFAULT_TEXT);
		exit (1);
	}

	/* GNU GPL suggests that we display this. */
	if (get_verbose() & VERBOSE_NORM) {
		printf("APRS Parser with Database Support, Copyright (C) 2008 James Jefferson Jarvis\n");
		printf("This software comes with ABSOLUTELY NO WARRANTY\n");
		printf("This is free software, and you are welcome to redistribute it ");
		printf("under certain conditions.\n");
	}

	if (get_verbose() & VERBOSE_NORM) {
		fprintf (stderr, "%sParser started%s\n", GREEN_TEXT, DEFAULT_TEXT);
	}

	/* Initialize start time marker, for packet/sec calculations. */
	time(&total_packet_count_marker);

	/* start processing input from internet server (if two additional arguments) or stdin */
	if ( optind + 2 == argc ) {
		inet_feed(argv[optind],atoi(argv[optind+1]),aUsername,aPassword,filter);
	} else {
		/* otherwise we are getting input from stdin */
		qa_hostid = set_qa_hostid( "stdin" );
		stdin_feed();
	}

		
	/* done parsing */

	/* Close database link so we are not hogging resources when we are done */
	close_connection();

	if (get_verbose() & VERBOSE_NORM)
		fprintf (stderr, "%sProgram terminated normally.%s\n", GREEN_TEXT, DEFAULT_TEXT);

	exit(0);
}


int valid_packet(char *buffer){
	// FIXME: is_valid should be used correctly.  No biggie, yet.
	int is_valid;

	// default to invalid
	is_valid = 0;

	// buffer must have characters else we segfault.
	if(buffer == NULL)
		return is_valid;

	// buffer should have more than a few characters
	if(strlen(buffer) > 2)  // FIXME: What is the shortest valid packet length?
		is_valid = 1;
	else
		return is_valid;

	// buffer should not be a server comment
	if (buffer[0] == '#') {
		is_valid = 0;
	}

	/* no archive users ... just ignore their packet, per their request */
	if ( NULL != strstr(buffer,"x-no-ar") || NULL != strstr(buffer,"!x!") ) {
		is_valid=0;
	}
	
	return is_valid;
}        


/**
 * Determines what parser to give each packet to 
 */
void process (char *buffer) {
	static int tick=0;
	unsigned long packet_id;
	struct raw_packet packet;
	struct message msg;
	struct position pos;
	struct status stat;
	struct weather wx;
	struct telemetry telem;
	struct object obj;
	struct item itm;

	strcpy (packet.source, "");
	strcpy (packet.destination, "");
	strcpy (packet.information, "");
	strcpy (packet.information, "");

        if(!valid_packet(buffer)){
		if (buffer[0] = '#') {
	                if( get_verbose() & VERBOSE_PARSE )
        	                printf("Message from server: %s\n", buffer);
		} else {
	                if( get_verbose() & VERBOSE_PARSE )
        	                printf("%sInvalid packet rejected: %s%s\n", RED_TEXT, buffer, DEFAULT_TEXT);
		}

                return;
        }

	/* Print the raw packet in case it gets mangled by parse_tnc */
	if ( get_verbose() & VERBOSE_DEBUG )
		printf("%s\n",buffer);

	/* Break the packet into the raw_packet structure, if we have a weird packet then we skip it */
	if ( 0 != parse_tnc (&packet, buffer) ) {
		if ( get_verbose() & VERBOSE_DEBUG ) 
			printf("Throwing out packet because parse_tnc reported it was corrupt\n");
		return;
	}

	/* check for destination address that is GPS?X? ... continue if we get it ... no archive */
	if ( NULL != strstr(packet.destination,"GPS") && strlen(packet.destination) > 5 && packet.destination[4]=='X') {
		if ( get_verbose() & VERBOSE_ERRORS )
			printf("GPS no-archive packet ... skipping\n");

		return;
	}

	init_message(&msg, packet.source);
	init_status(&stat, packet.source);
	init_position(&pos, packet.source);
	init_telemetry(&telem, packet.source);
	init_weather(&wx, packet.source);
	init_object(&obj, packet.source);
	init_item(&itm, packet.source);

  	/* if only_flag is set then we will skip anything that isn't from our station */
  	if ( only_flag && strcmp(packet.source,only_source) )
  			return;

	/* Save raw packet */
	packet_id = packet_insert (&packet);

	/* Source is common between them all */
	strcpy(msg.source, packet.source);
	strcpy(pos.source, packet.source);
	strcpy(stat.source, packet.source);
	strcpy(wx.source, packet.source);
	strcpy(telem.source, packet.source);
	strcpy(obj.source, packet.source);
	strcpy(itm.source, packet.source);

	/* We try determine the packet type based on the first character in the payload */
	switch (packet.information[0]) {
		case ':':		/* station to station message, new 1998 format */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Message                               %s\n", packet.source);

			if ( 0 == parse_message (&msg, packet.information) ) {
				/* Junk filter for USERLIST and javaMSG */
				message_insert (&msg, packet_id);
			}
			packet_count[MESSAGE]++;
			break;
		case '_':		/* weather */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Weather                               %s\n", packet.source);

			if ( parse_wx (&wx, packet.information) ) {
				weather_insert (&wx, packet_id);
			}
			packet_count[WEATHER]++;
			break;
		case '@':		/* APRS mobile station */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Mobile                                %s\n", packet.source);

			parse_position (&pos, &stat, packet.information, packet_id);
			position_insert (&pos, packet_id);
			packet_count[MOBILE]++;
			break;
		case '=':		/* APRS fixed statuin */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Fixed                                 %s\n", packet.source);

			parse_position (&pos, &stat, packet.information, packet_id);
			position_insert (&pos, packet_id);
			packet_count[FIXED]++;
			break;
		case '!':		/* APRS not runing, fixed short format, eburrows: position without timestamp*/

			if (get_verbose() & VERBOSE_DEBUG)
				printf("No-Timestamp                          %s\n", packet.source);

			if ( '!' == packet.information[1] ) {
				/* This is an ultimeter weather station in "data logging mode", 
				 * this should not be on the air
				 */
		                if (get_verbose() & VERBOSE_DEBUG)
			                printf ("Ultimeter Weather: data logging mode\n");
		                parse_peet_datalog(&wx, packet.information);
		                weather_insert (&wx, packet_id);
		                packet_count[WEATHER]++;				
			} else {	
				parse_position (&pos, &stat, packet.information, packet_id);
				position_insert (&pos, packet_id);
				packet_count[APRS_NOT_SHORT]++;
			}
			break;
		case '/':		/* APRS not running, fade to gray in 2 hrs, eburrows: position without timestamp */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("No-Timestamp                          %s\n", packet.source);

			parse_position (&pos, &stat, packet.information, packet_id);
			position_insert (&pos, packet_id);
			packet_count[APRS_NOT_FADE]++;
			break;
		case '>':		/* APRS status */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Status                                %s\n", packet.source);

			parse_status (&stat, &pos, packet.information, packet_id);	
			/* Some status packets contain position information as well. position_insert() checks to
			 * make sure the structure contains a valid packet before inserting it. Here is an
			 * example of a status packet that contains position data and will be inserted:
			 * KB0THN>APRS:>EN43EX Jim at home in Winona
			 */
			position_insert (&pos, packet_id);
			packet_count[STATUS]++;
			break;

		/* APRS MIC-E Packet */
		case 0x60:
		case 0x27:
		case 0x1c:
		case 0x1d:

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Mic-e                                 %s\n", packet.source);

			parse_mic_e (&pos, &stat, packet.information, packet.destination, packet_id);
			packet_count[MICE]++;
			break;


		case '$':		/* NMEA data, Ultimeter weather station in packet mode */
#if 0
			if ( strstr(packet.information,"$PNTS") ) {
			/* Navitra packets from Kenwood radio's in Japan */
				if (get_verbose() & VERBOSE_DEBUG)
					printf("NAVI-TRA Packet\n");

				/* change navitra packet to aprs */
				navitra_to_aprs(packet.information,packet.information);	
				parse_position (&pos, &stat, packet.information, packet_id);
				position_insert (&pos, packet_id);
				packet_count[NAVITRA]++;
			} else 
#endif				
			if ( strstr(packet.information,"$ULTW") ) { 
			/* This is an ultimeter weather station in "packet mode", it got here because it 
			 * looks like a NMEA sentence
			 */
				if (get_verbose() & VERBOSE_DEBUG)
					printf ("Ultimeter Weather: packet mode\n");
				parse_peet_packet(&wx, packet.information);
				weather_insert (&wx, packet_id);
				packet_count[WEATHER]++;	        
			} else {
			/* probably just a plain old NMEA packet */
				if (get_verbose() & VERBOSE_DEBUG)
					printf ("NMEA data\n");
				if ( 0 == parse_nmea (&pos, packet.information) )
					position_insert (&pos, packet_id);

				packet_count[NMEA]++;
			}
			break;

		/* Telemetry */
		case 'T':
			if (get_verbose() & VERBOSE_DEBUG)
				printf("Telemetry                             %s\n", packet.source);


			if ( '#' == packet.information[1] ) {
				if (get_verbose() & VERBOSE_DEBUG)
					printf ("Telemetry\n");
				if ( 0 == parse_telemetry(&telem, packet.information) )
					telemetry_insert (&telem, packet_id);
			}
			break;

		/* Not implemented */
		case '{':		/* User defined format, Bob B WB4APR is currently keeping the list */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("User defined                          %s\n%s\n", packet.source, packet.information);

			user_insert(packet.source,packet.information,packet_id);

			packet_count[USER]++;
			break;
		/* Not implemented */
		case '?':		/* APRS query */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Query from %s\n", packet.source);

			packet_count[QUERY]++;
			break;

		case ';':		/* APRS object */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Object                                %s\n", packet.source);

			packet_count[OBJECT]++;

			if (0 == parse_object(&obj, packet.information))
				object_insert(&obj, packet_id);

			break;

		case ')':		/* APRS item */

			if (get_verbose() & VERBOSE_DEBUG)
				printf("Item                                  %s\n", packet.source);

			if (0 == parse_item(&itm, packet.information))
				item_insert(&itm, packet_id);

			break;

		default:
			if (get_verbose() & VERBOSE_DEBUG)
				printf ("Unknown packet type\n");
			packet_count[UNKNOWN]++;
	}

	if ( insert_mask & INSERT_WATCH ) {
		if ( get_verbose() & VERBOSE_DEBUG )
			printf("Checking to see if \"%s\" is in watch_users\n",packet.source);

		watch_trigger(packet.source,packet_id);
		if ( msg.addressee[0] )	/* Also check the message addressee if it is not empty */
			watch_trigger(msg.addressee,packet_id);

		if ( NULL != strstr(packet.information,"EMERGENCY") ) {
			if ( get_verbose() & VERBOSE_DEBUG ) {
				printf("EMERGENCY contained in a packet\n");
			}
			watch_trigger("EMERG",packet_id);

		}
	}

	
	/* Last position table stores the only the most recent position information for each
	 * distinct callsign. */
	if ( pos.latitude != 0 && pos.longitude != 0 ) {
		if ( get_verbose() & VERBOSE_DEBUG)
			printf("Inserting last position\n");
		lastposition_insert (&pos, &stat, packet_id);
	}

	/* tick marks */
	tick++;
	if ( 0 == (tick % 100) ) {
		putchar('.');
		fflush(stdout);
		tick=0;
	}
	
	/* commit SQL transaction if needed */
	total_packet_count++;
	if (total_packet_count % packets_per_commit == 0) {
	
		if (get_verbose() & VERBOSE_NORM)
			printf("Commiting %d packets, %f packets/second.\n", packets_per_commit, (float)packets_per_commit / (float)(time(NULL) - total_packet_count_marker));

		db_commit_txn();
		db_begin_txn();

		time(&total_packet_count_marker);
	}


	/* Seperator bar */
	if (get_verbose() & VERBOSE_PARSE)
		printf ("/************************** NEXT PACKET **************************/\n");
}


void init_object(struct object *obj, char *source) {
	strncpy(obj->source,source,10);
	strncpy(obj->name, "\0\0\0\0\0\0\0\0\0\0", 10);
	obj->kill = 0;

	obj->wx.wind_direction=-1;
	obj->wx.wind_speed=-1;
	obj->wx.wind_gust=-1;
	obj->wx.wind_sustained=-1;
	obj->wx.temperature=-100;
	obj->wx.rain_hour=-1.0;
	obj->wx.rain_calendar_day=-1.0;
	obj->wx.rain_24hour_day=-1;
	obj->wx.humidity=0;
	obj->wx.barometer=-1;
	obj->wx.luminosity=-1;
	obj->wx.rain_count=0;

	obj->pos.latitude=0;
	obj->pos.longitude=0;
	obj->pos.speed_over_ground=-1;
	obj->pos.course_made_good=-1;
	obj->pos.altitude=0;

	obj->stat.comment[0] = '\0';
	obj->stat.symbol_table = ' ';
	obj->stat.symbol_code = ' ';
	obj->stat.power = 0;
	obj->stat.height = 0;
	obj->stat.gain = 0;
	obj->stat.directivity = 0;
	obj->stat.rate = 0;

}

int parse_compressed_object(struct object *obj, char *info) {
	int i, range;
	char *p = info;

	/* verify to see if our characters are in range */
	for ( i=1 ; i<9 ; i++ ) {
		if ( 9 == i || 12 == i )
			continue;

		if ( p[i] < '!' || p[i] > '{' ) {
//			printf("Compressed report contains invalid character at %d\n",i);
			return -1;
		}
		
	}

	obj->stat.symbol_table=p[0];
	obj->stat.symbol_code=p[9];
	strncpy(obj->stat.comment,&info[12],256);

	obj->pos.latitude = 90 - ( (p[1]-'!')*753571 + (p[2]-'!')*8281 + (p[3]-'!')*91 + (p[4]-'!') ) / 380926.0;
	obj->pos.longitude = -180 + ( (p[5]-'!')*753571 + (p[6]-'!')*8281 + (p[7]-'!')*91 + (p[8]-'!') ) / 190463.0;
	
	if ( ' ' != p[10] ) {
		if ( 16==(p[12]&24) ) {
			/* contains altitude information because bits 4,3 = 1,0 */
			obj->pos.altitude = (int) pow(1.002,( ((p[10]-'!')*91) + p[11]-'!'));
		} else if ( p[10] >= '!' && p[10] <= 'z' ) {
			/* course in degrees */
			obj->pos.course_made_good = (float) (p[10]-'!')*4;
			/* speed in knots */
			obj->pos.speed_over_ground = (float) NM_TO_KM(pow(1.08,(p[11]-'!')) - 1);
		} else if ( '{' == p[10] ) {
			//printf("radio range");
			/* pre-calculated radio range */
			range = (int) (2.0 * pow(1.09,(p[11]-'!')));
		}
	}
	
	/* bug: put the range into the obj->stat.comment field */
	
	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("  latitude=%lf\n",obj->pos.latitude);
		printf("  longitude=%lf\n",obj->pos.longitude);
		printf("  altitude=%d\n",obj->pos.altitude);
		printf("  symbol_table=%c\n",obj->stat.symbol_table);
		printf("  symbol_code=%c\n",obj->stat.symbol_code);
		printf("  course=%d\n",obj->pos.course_made_good);
		printf("  speed=%f\n",obj->pos.speed_over_ground);
		printf("  pre-calculated radio range=%d\n",range);
	}

	return 0;
}

int parse_compressed_item(struct item *itm, char *info) {
	int i, range;
	char *p = info;

	/* verify to see if our characters are in range */
	for ( i=1 ; i<9 ; i++ ) {
		if ( 9 == i || 12 == i )
			continue;

		if ( p[i] < '!' || p[i] > '{' ) {
			return -1;
		}
		
	}

	itm->stat.symbol_table=p[0];
	itm->stat.symbol_code=p[9];
	strncpy(itm->stat.comment,&info[12],256);

	itm->pos.latitude = 90 - ( (p[1]-'!')*753571 + (p[2]-'!')*8281 + (p[3]-'!')*91 + (p[4]-'!') ) / 380926.0;
	itm->pos.longitude = -180 + ( (p[5]-'!')*753571 + (p[6]-'!')*8281 + (p[7]-'!')*91 + (p[8]-'!') ) / 190463.0;
	
	if ( ' ' != p[10] ) {
		if ( 16==(p[12]&24) ) {
			/* contains altitude information because bits 4,3 = 1,0 */
			itm->pos.altitude = (int) pow(1.002,( ((p[10]-'!')*91) + p[11]-'!'));
		} else if ( p[10] >= '!' && p[10] <= 'z' ) {
			/* course in degrees */
			itm->pos.course_made_good = (float) (p[10]-'!')*4;
			/* speed in knots */
			itm->pos.speed_over_ground = (float) NM_TO_KM(pow(1.08,(p[11]-'!')) - 1);
		} else if ( '{' == p[10] ) {
			//printf("radio range");
			/* pre-calculated radio range */
			range = (int) (2.0 * pow(1.09,(p[11]-'!')));
		}
	}
	
	/* bug: put the range into the itm->stat.comment field */
	
	if ( get_verbose() & VERBOSE_DEBUG ) {
		printf("  latitude=%lf\n",itm->pos.latitude);
		printf("  longitude=%lf\n",itm->pos.longitude);
		printf("  altitude=%d\n",itm->pos.altitude);
		printf("  symbol_table=%c\n",itm->stat.symbol_table);
		printf("  symbol_code=%c\n",itm->stat.symbol_code);
		printf("  course=%d\n",itm->pos.course_made_good);
		printf("  speed=%f\n",itm->pos.speed_over_ground);
		printf("  pre-calculated radio range=%d\n",range);
	}

	return 0;
}

// From Xastir 2004/02/15
// Extract single weather data item from "data".  Returns it in
// "temp".  Modifies "data" to remove the found data from the
// string.  Returns a 1 if found, 0 if not found.
//
// PE1DNN
// If the item is contained in the string but does not contain a
// value then regard the item as "not found" in the weather string.
//
int extract_weather_item(char *data, char type, int datalen, char *temp) {
	int i,ofs,found,len;

	found=0;
	len = (int)strlen(data);

	if (data[0] == type && strlen(data) + 1 >= datalen) {

		found = 1;

		for(i = 1; i < datalen; ++i) {
			if (isdigit(data[i]) == 0 && data[i] != '.' && data[i] != 'V') {
				found = 0;
				break;
			}

			temp[i-1] = data[i];
		}

		temp[datalen] = '\0';
	}

	return(found);
}

int parse_object(struct object *obj, char *p) {
	char name[10];
	char *name_space;
	char buffer[256];

	/* test to see if we actually have anything */
	if (strlen(p) >= 10) {
		/* name can be up to 10 characters */
		strncpy(name, p + 1, 9);
		name[9]='\0';

		if ((name_space = strchr(name, ' ')) != NULL) {
			name[name_space - name]='\0';
			strncpy(obj->name, name, name_space - name);
		} else {
			strncpy(obj->name, name, 9);
		}
	} else {
		return -1;
	}

	/* compressed reports will start with non-digit */
	if ( ! isdigit(*(p+12))) {
		return parse_compressed_object(obj, p);
	} 

	if (p[10] == '_') {
		obj->kill = 1;
	}

	p = p + 18; // Jump past the time to the latitude

	/* latitude */
	strncpy (buffer, p, 7);	
	buffer[7] = '\0';

	obj->pos.latitude = dm2dd(atof (buffer) / 100);
	p += 7;
	if ('S' == *p)
		obj->pos.latitude = 0 - obj->pos.latitude;
	p++;

	/* symbol table */
	obj->stat.symbol_table = *p;	
	p++;

	/* longitude */
	strncpy (buffer, p, 8);	// Longitude
	buffer[8] = '\0';
	obj->pos.longitude = dm2dd(atof (buffer) / 100);
	p += 8;
	if ('W' == *p)
		obj->pos.longitude = 0 - obj->pos.longitude;
	p++;

	/* symbol code */
	obj->stat.symbol_code = *p;
	p++;

	// Check if we have the  123/987	course/speed information
	if (  isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2]) && '/' == p[3] && 
	isdigit(p[4]) && isdigit(p[5]) && isdigit(p[6]) ) {
		memcpy(buffer,p,3);
		buffer[3] = '\0';
		obj->pos.course_made_good = (float) atoi(buffer);
		memcpy(buffer,p+4,3);
		buffer[3] = '\0';
		obj->pos.speed_over_ground = NM_TO_KM((float) atoi(buffer));
		p +=7;	// Swallow this
	}

	if (get_verbose() & VERBOSE_DEBUG)
		printf ("Parse object status started with: %s\n", p);


	extract_powergain (p, &obj->stat);

	if (0 != extract_altitude (p, buffer)) {
		obj->pos.altitude = FT_TO_M(atof (buffer));

		if (get_verbose() & VERBOSE_DEBUG)
			printf ("Altiude = %f\n", obj->pos.altitude);
	}


	/*
	 * From Xastir project, db.c 2004/02/15
	 */


	// gust (peak wind speed in mph in the last 5 minutes)
	if (extract_weather_item(p,'g',3, buffer) == 1) {     
		obj->wx.wind_gust = atoi(buffer);
		p += 4;
	}
	
	// temperature (in deg Fahrenheit), could be negative
	if (extract_weather_item(p,'t',3, buffer) == 1) {      
		obj->wx.temperature = F_TO_C(atoi(buffer));
		p += 4;
	}

	// rainfall (1/100 inch) in the last hour
	if (extract_weather_item(p,'r',3, buffer) == 1) {     
		obj->wx.rain_hour = atof(buffer);
		p += 4;
	}

	// rainfall (1/100 inch) in the last 24 hours
	if (extract_weather_item(p,'p',3, buffer) == 1) {   
		obj->wx.rain_24hour_day = atof(buffer);
		p += 4;
	}

	// rainfall (1/100 inch) since midnight
	if (extract_weather_item(p,'P',3, buffer) == 1) {   
		obj->wx.rain_calendar_day = atof(buffer);
		p += 4;
	}

	// snowfall (in inches) in the last 24 hours
	// was 1/100 inch, APRS reference says inch! ??
	if (extract_weather_item(p,'s',3, buffer) == 1) {      
		obj->wx.snowfall_24hour_day = atof(buffer);
		p += 4;
	}

	// luminosity (in watts per square meter) 999 and below
	if (extract_weather_item(p,'L',3, buffer) == 1) {                  
		obj->wx.luminosity = atoi(buffer);
		p += 4;
	}

	// luminosity (in watts per square meter) 1000 and above
	if (extract_weather_item(p,'l',3, buffer) == 1) {                  
		obj->wx.luminosity = atoi(buffer);
		p += 4;
	}

	// raw rain counter
	if (extract_weather_item(p,'#',3, buffer) == 1) {
		obj->wx.rain_count = atoi(buffer);
		p += 4;
	}

	// Fuel Temperature in deg F (RAWS)
   if (extract_weather_item(p,'F',3, buffer) == 1) { 
		obj->wx.fuel_temp = atoi(buffer);
		p += 4;
	}


	// humidity (in %, 00 = 100%)
	if (extract_weather_item(p,'h',2, buffer)) {         
		obj->wx.humidity = atoi(buffer);
		p += 3;
	}

	// barometric pressure (1/10 mbar / 1/10 hPascal)
	if (extract_weather_item(p,'b',5, buffer)) {  
		obj->wx.barometer = atof(buffer);
		p += 6;
	}

	// Fuel Moisture (RAWS) (in %, 00 = 100%)
	if (extract_weather_item(p,'f',2, buffer)) { 
		obj->wx.fuel_moisture = atof(buffer);
		p += 3;
	}


	/* If the object is sending weather data, convert speed/course into wind speed/course */
	if (obj->wx.temperature != -100 || obj->wx.wind_gust != -1 || obj->wx.rain_hour != -1.0) {
		obj->wx.wind_direction = obj->pos.course_made_good;
		obj->wx.wind_speed = obj->pos.speed_over_ground;

		obj->pos.course_made_good = -1;
		obj->pos.speed_over_ground = 0;
	}

	strncpy(obj->stat.comment, p, 256);


	if ( get_verbose() & VERBOSE_DEBUG) {
		printf ("  source: '%s'\n", obj->source);
		printf ("  name: '%s'\n", obj->name);
		printf ("  latitude: %f\n", obj->pos.latitude);
		printf ("  symbol table: %c\n", obj->stat.symbol_table);
		printf ("  longitude: %f\n", obj->pos.longitude);
		printf ("  symbol code: %c\n", obj->stat.symbol_code);
		printf ("  course: %f\n", obj->pos.course_made_good);
		printf ("  speed: %f\n", obj->pos.speed_over_ground);
		printf ("  altitude %f\n", obj->pos.altitude);
		printf ("  power %d\n", obj->stat.power);
		printf ("  height %d\n", obj->stat.height);
		printf ("  gain %f\n", obj->stat.gain);
		printf ("  directivity %d\n", obj->stat.directivity);
		printf ("  rate %d\n", obj->stat.rate);
		printf ("  comment: %s\n", obj->stat.comment);
	}

	return 0;
}

/**
 * Initialize telemetry structure
 */
void init_telemetry(struct telemetry *telem, char *source) {
	strncpy(telem->source,source,10);

	telem->sequence=0;
	telem->analog[0]=0;
	telem->analog[1]=0;
	telem->analog[2]=0;
	telem->analog[3]=0;
	telem->analog[4]=0;
	telem->digital=0;
}

/**
 * Initialize weather structure
 */
void init_weather(struct weather *wx, char *source) {
	strncpy(wx->source,source,10);
	wx->wind_direction=-1;
	wx->wind_speed=-1;
	wx->wind_gust=-1;
	wx->wind_sustained=-1;
	wx->temperature=-100;
	
	wx->rain_hour=-1.0;
	wx->rain_calendar_day=-1.0;
	wx->rain_24hour_day=-1;

	wx->humidity=0;
	wx->barometer=-1;
	wx->luminosity=-1;
	wx->rain_count=0;

	wx->snowfall_24hour_day=0;
	wx->fuel_temp=-100;
	wx->fuel_moisture=0;
}

/**
 * Initialize message structure
 */
void init_message(struct message *msg, char *source) {
	strncpy(msg->source,source,10);
	strcpy(msg->addressee,"");
	strcpy(msg->text,"");
	strcpy(msg->id,"");
}

/**
 * Initialize status structure
 */
void init_status(struct status *stat, char *source) {
	strncpy(stat->source,source,10);

	stat->comment[0] = '\0';
	stat->power=0;
	stat->height=0;
	stat->gain=0;
	stat->directivity=-1;
	stat->rate=0;
	stat->symbol_table = '\0';
	stat->symbol_code = '\0';
}

/**
 * Initialize position structure
 */
void init_position(struct position *pos, char *source) {
	strncpy(pos->source,source,10);
	pos->time_of_fix=0;
	pos->latitude=0;
	pos->longitude=0;
	pos->speed_over_ground=-1;
	pos->course_made_good=-1;
	pos->altitude=0;

	pos->checksum=0;
	pos->sentence[0]='\0';
	pos->rx_warning=0;
	pos->magnetic_variance=0;
	pos->quality=0;
	pos->sats=0;
	pos->hor_dilution=0;
	pos->geoidal_separation=0;
	pos->diff_last_update=0;
	pos->diff_station=0;
}

void init_item(struct item *itm, char *source) {
	strncpy(itm->source,source,10);
	strncpy(itm->name, "\0\0\0\0\0\0\0\0\0\0", 10);
	itm->kill = 0;

	itm->pos.latitude=0;
	itm->pos.longitude=0;
	itm->pos.speed_over_ground=-1;
	itm->pos.course_made_good=-1;
	itm->pos.altitude=0;

	itm->stat.comment[0] = '\0';
	itm->stat.symbol_table = ' ';
	itm->stat.symbol_code = ' ';
	itm->stat.power = 0;
	itm->stat.height = 0;
	itm->stat.gain = 0;
	itm->stat.directivity = 0;
	itm->stat.rate = 0;

}

int parse_item(struct item *itm, char *p) {
	char buffer[256];
	int i;

	// Search for the name termination characters
	for( i = 1; i < 11; ++i) {
		if (p[i] == '!' || p[i] == '_') {
			break;
		}
	}

	strncpy(itm->name, p+1, i - 1);
	p += i;


	if (p[0] == '_') {
		itm->kill = 1;
	}

	p++; // Advance past seperator


	/* compressed reports will start with non-digit */
	if ( ! isdigit(p[0])) {
		return parse_compressed_item(itm, p);
	} 

	/* latitude */
	strncpy (buffer, p, 7);	
	buffer[7] = '\0';

	itm->pos.latitude = dm2dd(atof (buffer) / 100);
	p += 7;
	if ('S' == *p)
		itm->pos.latitude = 0 - itm->pos.latitude;
	p++;

	/* symbol table */
	itm->stat.symbol_table = *p;	
	p++;

	/* longitude */
	strncpy (buffer, p, 8);	// Longitude
	buffer[8] = '\0';
	itm->pos.longitude = dm2dd(atof (buffer) / 100);
	p += 8;
	if ('W' == *p)
		itm->pos.longitude = 0 - itm->pos.longitude;
	p++;

	/* symbol code */
	itm->stat.symbol_code = *p;
	p++;

	// Check if we have the  123/987	course/speed information
	if (  isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2]) && '/' == p[3] && 
	isdigit(p[4]) && isdigit(p[5]) && isdigit(p[6]) ) {
		memcpy(buffer,p,3);
		buffer[3] = '\0';
		itm->pos.course_made_good = (float) atoi(buffer);
		memcpy(buffer,p+4,3);
		buffer[3] = '\0';
		itm->pos.speed_over_ground = NM_TO_KM((float) atoi(buffer));
		p +=7;	// Swallow this
	}

	if (get_verbose() & VERBOSE_DEBUG)
		printf ("Parse item status started with: %s\n", p);


	extract_powergain (p, &itm->stat);


	// Stuff the rest into the comment field
	strcpy(itm->stat.comment, p);

	if (get_verbose() & VERBOSE_DEBUG) {
		printf ("  source: '%s'\n", itm->source);
		printf ("  name: '%s'\n", itm->name);
		printf ("  latitude: %f\n", itm->pos.latitude);
		printf ("  symbol table: %c\n", itm->stat.symbol_table);
		printf ("  longitude: %f\n", itm->pos.longitude);
		printf ("  symbol code: %c\n", itm->stat.symbol_code);
		printf ("  power %d\n", itm->stat.power);
		printf ("  height %d\n", itm->stat.height);
		printf ("  gain %f\n", itm->stat.gain);
		printf ("  directivity %d\n", itm->stat.directivity);
		printf ("  rate %d\n", itm->stat.rate);
		printf ("  comment: %s\n", itm->stat.comment);
	}

	return 0;
}
