// vim:ts=3:
#ifndef _MESSAGE_H

#define BUFFSIZE	4096

unsigned int qa_hostid; // Host ID we should use when inserting raw data
unsigned int packet_count[20];  /* Counter of each type of packet */
#define MESSAGE			0	// 8    DONE
#define WEATHER			1	// 7    DONE
#define MOBILE				2	// 3    DONE
#define FIXED				3	// 2    DONE
#define APRS_NOT_SHORT	4	// 1    DONE
#define APRS_NOT_FADE	5	// 9    DONE
#define STATUS				6	// 6    DONE
#define QUERY				7	// 11
#define OBJECT				8	// 10
#define MICE				9	// 4    DONE - converts to standard APRS
#define USER				10	// 12   won't bother    - log it
#define NMEA				11	// 5    GPGAA,GPRMC DONE
#define UNKNOWN			12	// 7  
#define NAVITRA			13
#define TELEMETRY 		14

#define RED_TEXT "\033[31m"	// VT100 escape codes for changing color - prolly only for interactive
#define GREEN_TEXT "\033[32m"
#define DEFAULT_TEXT "\033[0m"

#define INSERT_RAW      1
#define INSERT_MESSAGE  2
#define INSERT_POSITION 4
#define INSERT_STATUS   8
#define INSERT_WEATHER  16
#define INSERT_LAST     32 /* all last tables {laststatus, lastweather, lastposition, lastuserdefined } */
#define INSERT_WATCH    64
#define INSERT_USER     128
#define INSERT_TELEM    256

#define VERBOSE_ERRORS		1	/* system errors */
#define VERBOSE_PARSE		2	/* packet parse problems */
#define VERBOSE_DEBUG		4	/* debugging information */
#define VERBOSE_NORM			8  /* Normal logging */
#define VERBOSE_SQL			16 /* SQL statements */


struct telemetry {
	char source[10];	// AX.25 Source address
	int sequence;	// Sequence no
	int analog[5];	// analog channels
	int digital;
};

void init_telemetry(struct telemetry *telem, char *source);

struct weather {
	char source[10];	// AX.25 Source address
	int wind_direction;
	int wind_speed;	// kilometers / hour
	int wind_gust;	// kilometers / hour
	int wind_sustained;	// kilometers / hour
	int temperature;	// celcius
	float rain_hour;	// centimeters
	float rain_calendar_day;	// centimeters
	float rain_24hour_day;	// centimeters
	int rain_counter;
	int humidity;
	float barometer;	// milibars
	int luminosity;
	int rain_count;
	int snowfall_24hour_day;
	int fuel_temp;
	int fuel_moisture;
};

void init_weather(struct weather *wx, char *source);

struct raw_packet {
	unsigned long packet_id;	// System wide unqiue packet ID
	// use MYSQL's last_insert_id()
	time_t time;		// Seconds since Epoch (00:00:00 UTC, January
	// 1, 1970)  
	// use MYSQL's from_unixtime() to convert
	// and MYSQL's unix_timestamp()
	char ip_source[255];	// IGate station's IP
	char source[10];	// AX.25 Source address
	char destination[11];	// AX.25 Destination address
	char digipeater[8][11];	// Digipeater addresses
	char information[512];	// Information field (spec says 256 bytes)
};

/* parse_tnc provides the initialization required */

/* station to station messages */
struct message {
	char source[10];	// AX.25 Source address
	char addressee[10];	// APRS message addressee
	char text[68];	// APRS message body
	char id[6];		// APRS message ID
};

void init_message(struct message *msg, char *source);

/* status information */
struct status {
	char source[10];	// AX.25 Source address
	char comment[256];	// Comment
	int power;		// Output power in Watts
	int height;		// Height above terain in meters
	float gain;		// antenna gain
	int directivity;	// antenna directivity
	int rate;		// station transmit rate
	unsigned char symbol_table;	// Symbol table
	unsigned char symbol_code;	// Symbol code  
};

void init_status(struct status *stat, char *source);

/* station position */
struct position {
	char source[10];	// AX.25 Source address
	time_t time_of_fix;	// Time of Fix (seconds since Epoch)
	double latitude;	// Latitude (North=postive,South=negative)
	double longitude;	// Longitude (West=negative,East=positive)
	float speed_over_ground;	// Speed over ground (km/hr)
	float course_made_good;	// Course made good (true)
	float altitude;	// Antenna altitude above / below sea level
	// APRS stuff
	// Less useful stuff (specific to instrument) below:
	int checksum;		// Checksum (Valid=1,Invalid=0)
	char sentence[6];	// Sentence type
	// BUG: Should be enumerated
	char rx_warning;	// Navigation receiver warning
	float magnetic_variance;	// Magnetic variance
	int quality;		// 0 = fix not available
							// 1 = GPS SPS mode
							// 2 = diffrential GPS, SPS mode
							// 3 = GPS PPS mode
	int sats;		// Satellites in use
	float hor_dilution;	// Horizontal dillution of precision
	float geoidal_separation;	// difference between teh WGS84 earth
	// ellipsoid and mean sea level
	float diff_last_update;	// seconds since last DGPS update
	int diff_station;	// diffrential station ID (0000-1023)
};

void init_position(struct position *pos, char *source);


struct object {
	char source[10];
	char name[10];
	int kill;

	struct weather wx;
	struct position pos;
	struct status stat;
};

void init_object(struct object *obj, char *source);


struct item {
	char source[10];
	char name[10];
	int kill;

	struct position pos;
	struct status stat;
};

void init_item(struct item *obj, char *source);


typedef struct {
	   char hostname[256];
	   int port;
} Inet_server;



/* Function prototypes for useful functions */
void substr (char *dest, char *src, int size);

/* Prototypes for shared functions */
void laststatus_insert (struct status *stat, unsigned long packet_id);
unsigned int get_insert_information(void);
unsigned int get_verbose(void);
int get_do_last_tables(void);
int set_qa_hostid( char *hostname );

#define _MESSAGE_H
#endif
