/*
mysql to shapefile export program
Copyright (C) 2002 James Jefferson Jarvis <jjeffers@aprsworld.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <mysql.h>
#include "shapefil.h"

double atof();	

/* Mysql stuff */
MYSQL_RES *result;
MYSQL_ROW row;
MYSQL mysql;
MYSQL_FIELD *field;

/* icky globals */
int verbose=0;


int make_sql_connection (char *host, char *username, char *password, char *database) {
	mysql_init (&mysql);
	
	if (!mysql_connect (&mysql, host, username, password)) {
		fprintf (stderr, "Failed to connect to database:\n%s\n", mysql_error (&mysql));
		return 1;
	}
	
	if (mysql_select_db (&mysql, database)) {
		fprintf (stderr, "Failed to select database:\n%s\n", mysql_error (&mysql));
		mysql_close (&mysql);
		return 2;
	}
	
	return 0;
}

int main(int argc, char *argv[]) {

	/* mysql paramaters */
	char mysql_database[256], mysql_host[256], mysql_user[256], mysql_pass[256];
	
	char sql[2048];
	int i,n;
	int shape=0;

	/* shapefile stuff */
	SHPHandle  hSHP;
	DBFHandle  hDBF;
	SHPObject  *psObject;
	double x,y,z;

	int query_fields;

	/* Default database values */
	strcpy (mysql_database, "aprs");
	strcpy (mysql_host, "db.aprsworld.net");
	strcpy (mysql_user, "guest");
	strcpy (mysql_pass, "");

	/* command line options */
	while ((n = getopt (argc, argv, "h:d:u:p:v")) != -1) {
		switch (n) {
			case 'h':
				strncpy(mysql_host,optarg,256);
				break;
			case 'd':
				strncpy(mysql_database,optarg,256);
				break;
			case 'u':
				strncpy(mysql_user,optarg,256);
				break;
			case 'p':
				strncpy(mysql_pass,optarg,256);
				break;
			case 'v':
				verbose=1;
				break;
		}
	}

	if ( optind+2 != argc ) {
		fprintf(stderr,"usage: %s [OPTIONS] shapefile sql_query\n",argv[0]);
		fprintf(stderr," -h mysql server\n");
		fprintf(stderr," -d mysql database\n");
		fprintf(stderr," -u mysql username\n");
		fprintf(stderr," -p mysql password\n");
		fprintf(stderr," -v verbose\n");
		exit (2);
	}

	/* Connect to MySQL database */	   
	if (make_sql_connection(mysql_host, mysql_user, mysql_pass, mysql_database)) {
		fprintf (stderr, "Could not connect to MySQL database\n");
		exit (1);
	}

	/* get SQL query */
	strncpy(sql,argv[optind+1],2048);

	hSHP = SHPCreate( argv[optind], SHPT_POINT );
	hDBF = DBFCreate( argv[optind] );
	if ( NULL == hSHP || NULL == hDBF ) {
		fprintf(stderr,"Unable to create shapefile\n");
		return -4;
	}
	
	if ( verbose ) {	
		printf("SQL Query: %s\n",sql);
	}

	if ( mysql_query(&mysql,sql) ) {
		fprintf(stderr,"MySQL query failed\n");;
		fprintf(stderr,"%s\n",mysql_error(&mysql));
		exit (3);
	}
	
	result = mysql_use_result (&mysql);
	query_fields = mysql_num_fields(result);

	if ( verbose ) {
		printf("Query is %d fields wide\n",query_fields);
	}
		
	/* determine field names for the DBF file
	 * y,x,z,dbf_field0,dbf_field1,dbf_fieldN 
	 */
	for(i = 3; i < query_fields; i++) {
		field = mysql_fetch_field_direct(result, i);
	
		if ( verbose ) {
			printf("Adding field '%s' to dbf\n", field->name);
		}

		DBFAddField( hDBF, field->name, FTString, 32, 0 );	/* all fields are 32 character strings */
	}

	/* add shapefile record while we have records available */
	n=0;
	while ( (row = mysql_fetch_row(result)) ) {
		if ( verbose ) {
			printf("[%d] ",n);
		}

		n++;

		/* skip records where x or y coordinate is null */
		if ( ! row[0] || ! row[1] ) {
			if ( verbose ) {
				printf("Skipping record...\n");
			}
			continue;
		}
		x = atof(row[0]);
		y = atof(row[1]);

		/* if z is null then we use 0 */
		if ( row[2] ) {
			z = atof(row[2]);
		} else {
			if ( verbose ) {
				printf("No z-coordinate information... ");
			}
			z = 0;
		}

		/* create and write the shape object */
		psObject = SHPCreateSimpleObject(SHPT_POINT,1,&x,&y,&z);	/* type, verticies, x, y, z */
		SHPWriteObject( hSHP, -1, psObject );
		SHPDestroyObject( psObject );

		/* write the fields after x,y,z to the DBF file */
		for ( i=3 ; i<query_fields ; i++ ) {
			if ( ! row[i] )
				continue;
			DBFWriteStringAttribute( hDBF, shape, i-3, row[i] );
		}

		if ( verbose ) {
			printf("Lat: %lf Lon: %lf Altitude: %lf\n",y,x,z);
		}

		shape++;
	}

	/* close database connections and free memory */
	mysql_free_result(result);
	mysql_close(&mysql);

	SHPClose( hSHP );
	DBFClose( hDBF );
	return 0;
}
